import json
from copy import deepcopy

from django.test import TestCase, TransactionTestCase
from django.urls import reverse
from django.utils.timezone import now
from swapper import load_model

from openwisp_controller.config.tests.utils import CreateConfigTemplateMixin

from ...monitoring.tests import TestMonitoringMixin
from ..utils import manage_short_retention_policy

Metric = load_model('monitoring', 'Metric')
DeviceData = load_model('device_monitoring', 'DeviceData')
WifiClient = load_model('device_monitoring', 'WifiClient')
WifiSession = load_model('device_monitoring', 'WifiSession')
Chart = load_model('monitoring', 'Chart')
Config = load_model('config', 'Config')
Device = load_model('config', 'Device')


class TestDeviceMonitoringMixin(CreateConfigTemplateMixin, TestMonitoringMixin):
    device_model = Device
    config_model = Config
    _PING = 'openwisp_monitoring.check.classes.Ping'

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        manage_short_retention_policy()

    def _url(self, pk, key=None, time=None):
        url = reverse('monitoring:api_device_metric', args=[pk])
        if key:
            url = '{0}?key={1}'.format(url, key)
        if time:
            url = '{0}&time={1}'.format(url, time)
        return url

    def _post_data(self, id, key, data, time=None):
        if not time:
            time = now().utcnow().strftime('%d-%m-%Y_%H:%M:%S.%f')
        url = self._url(id, key, time)
        netjson = json.dumps(data)
        return self.client.post(url, netjson, content_type='application/json')

    def _create_device_monitoring(self):
        d = self._create_device(organization=self._create_org())
        dm = d.monitoring
        dm.status = 'ok'
        dm.save()
        return dm

    def _transform_wireless_interface_test_data(self, data):
        for interface in data.get('interfaces', []):
            wireless = interface.get('wireless')
            if wireless and all(key in wireless for key in ('htmode', 'clients')):
                for client in wireless['clients']:
                    htmode = wireless['htmode']
                    ht_enabled = htmode.startswith('HT')
                    vht_enabled = htmode.startswith('VHT')
                    noht_enabled = htmode == 'NOHT'
                    if noht_enabled:
                        client['ht'] = client['vht'] = None
                        # since 'he' field is optional
                        if 'he' in client:
                            client['he'] = None
                    elif ht_enabled:
                        if client['vht'] is False:
                            client['vht'] = None
                        if client.get('he') is False:
                            client['he'] = None
                    elif vht_enabled and client.get('he') is False:
                        client['he'] = None
            if wireless and 'bitrate' in wireless:
                wireless['bitrate'] = round(wireless['bitrate'] / 1000.0, 1)
        return data

    def assertDataDict(self, dd_data, data):
        """Compares monitoring data.

        This method is necessary because the wireless interface data is
        modified by the `AbstractDeviceData._transform_data` method.
        """
        data = self._transform_wireless_interface_test_data(data)
        self.assertDictEqual(dd_data, data)

    def create_test_data(self, no_resources=False, data=None, assertions=True):
        o = self._create_org()
        d = self._create_device(organization=o)
        if not data:
            data = self._data()
        # creation of resources metrics can be avoided in tests not involving them
        # this speeds up those tests by reducing requests made
        if no_resources:
            del data['resources']
        r = self._post_data(d.id, d.key, data)
        self.assertEqual(r.status_code, 200)
        dd = DeviceData(pk=d.pk)
        if not assertions:
            return dd
        self.assertDataDict(dd.data, data)
        if no_resources:
            metric_count, chart_count = 4, 4
        else:
            metric_count, chart_count = 7, 7
        # Exclude general metrics from the query
        self.assertEqual(Metric.objects.exclude(object_id=None).count(), metric_count)
        # Exclude general charts from the query
        self.assertEqual(
            Chart.objects.exclude(metric__object_id=None).count(), chart_count
        )
        if_dict = {'wlan0': data['interfaces'][0], 'wlan1': data['interfaces'][1]}
        extra_tags = {'organization_id': str(d.organization_id)}
        for ifname in ['wlan0', 'wlan1']:
            iface = if_dict[ifname]
            m = Metric.objects.get(
                key='traffic',
                field_name='rx_bytes',
                object_id=d.pk,
                main_tags={'ifname': ifname},
                extra_tags=extra_tags,
            )
            points = self._read_metric(
                m, limit=10, order='-time', extra_fields=['tx_bytes']
            )
            self.assertEqual(len(points), 1)
            for field in ['rx_bytes', 'tx_bytes']:
                self.assertEqual(points[0][field], iface['statistics'][field])
            m = Metric.objects.get(
                key='wifi_clients',
                field_name='clients',
                object_id=d.pk,
                extra_tags=extra_tags,
                main_tags={'ifname': ifname},
            )
            points = self._read_metric(m, limit=10, order='-time')
            self.assertEqual(len(points), len(iface['wireless']['clients']))
        return dd

    def _create_multiple_measurements(self, create=True, no_resources=False, count=4):
        if create:
            self.create_test_data(no_resources=no_resources)
        self.assertEqual(self.device_model.objects.count(), 1)
        d = self.device_model.objects.first()
        dd = DeviceData(pk=d.pk)
        data = self._data()
        # creation of resources metrics can be avoided in tests not involving them
        # this speeds up those tests by reducing requests made
        if no_resources:
            del data['resources']
        data2 = deepcopy(data)
        data2['interfaces'][0]['statistics']['rx_bytes'] = 400000000
        data2['interfaces'][0]['statistics']['tx_bytes'] = 100000000
        data2['interfaces'][1]['statistics']['rx_bytes'] = 2000000000
        data2['interfaces'][1]['statistics']['tx_bytes'] = 1000000000
        r = self._post_data(d.id, d.key, data2)
        if count == 2:
            return dd
        data3 = deepcopy(data)
        data3['interfaces'][0]['statistics']['rx_bytes'] = 500000000
        data3['interfaces'][0]['statistics']['tx_bytes'] = 300000000
        data3['interfaces'][1]['statistics']['rx_bytes'] = 0
        data3['interfaces'][1]['statistics']['tx_bytes'] = 0
        r = self._post_data(d.id, d.key, data3)
        if count == 3:
            return dd
        data4 = deepcopy(data)
        data4['interfaces'][0]['statistics']['rx_bytes'] = 1200000000
        data4['interfaces'][0]['statistics']['tx_bytes'] = 600000000
        data4['interfaces'][1]['statistics']['rx_bytes'] = 1000000000
        data4['interfaces'][1]['statistics']['tx_bytes'] = 500000000
        r = self._post_data(d.id, d.key, data4)
        self.assertEqual(r.status_code, 200)
        self.assertDataDict(dd.data, data4)
        return dd

    def _data(self):
        return {
            'type': 'DeviceMonitoring',
            'general': {'local_time': 1589026500, 'uptime': 8003},
            'resources': {
                'cpus': 1,
                'memory': {
                    'total': 249774080,
                    'shared': 86016,
                    'free': 224497664,
                    'cached': 6774784,
                    'available': 223397664,
                    'buffered': 974848,
                },
                'load': [0, 0, 0],
                'disk': [
                    {
                        'used_bytes': 18792,
                        'available_bytes': 233984,
                        'filesystem': '/dev/root',
                        'mount_point': '/',
                        'used_percent': 7,
                        'size_bytes': 258016,
                    },
                    {
                        'used_bytes': 3872,
                        'available_bytes': 11916,
                        'filesystem': '/dev/sda1',
                        'mount_point': '/boot',
                        'used_percent': 25,
                        'size_bytes': 16112,
                    },
                ],
                'swap': {'free': 0, 'total': 0},
            },
            'interfaces': [
                {
                    'name': 'wlan0',
                    'type': 'wireless',
                    'up': True,
                    'mac': '44:d1:fa:4b:38:43',
                    'txqueuelen': 1000,
                    'multicast': True,
                    'mtu': 1500,
                    'statistics': {
                        'rx_bytes': 324,
                        'tx_bytes': 145,
                        'collisions': 0,
                        'multicast': 0,
                        'rx_dropped': 0,
                        'tx_dropped': 0,
                    },
                    'wireless': {
                        'frequency': 2437,
                        'mode': 'access_point',
                        'signal': -29,
                        'tx_power': 6,
                        'channel': 6,
                        'ssid': 'testnet',
                        'noise': -95,
                        'country': 'US',
                        'htmode': 'HT20',
                        'clients': [
                            {
                                'mac': '00:ee:ad:34:f5:3b',
                                'wps': False,
                                'wds': False,
                                'ht': True,
                                'preauth': False,
                                'assoc': True,
                                'authorized': True,
                                'vht': False,
                                'he': False,
                                'wmm': True,
                                'aid': 1,
                                'mfp': False,
                                'auth': True,
                                'signature': 'test_signature',
                            }
                        ],
                    },
                },
                {
                    'name': 'wlan1',
                    'type': 'wireless',
                    'up': True,
                    'mac': '44:d1:fa:4b:38:44',
                    'txqueuelen': 1000,
                    'multicast': True,
                    'mtu': 1500,
                    'statistics': {
                        'rx_bytes': 2275,
                        'tx_bytes': 826,
                        'collisions': 0,
                        'multicast': 0,
                        'rx_dropped': 0,
                        'tx_dropped': 0,
                    },
                    'wireless': {
                        'frequency': 2437,
                        'mode': 'access_point',
                        'signal': -29,
                        'tx_power': 6,
                        'channel': 6,
                        'ssid': 'testnet',
                        'bitrate': 1100,
                        'quality': 65,
                        'quality_max': 70,
                        'noise': -95,
                        'country': 'US',
                        'htmode': 'VHT80',
                        'clients': [
                            {
                                'mac': 'b0:e1:7e:30:16:44',
                                'wps': False,
                                'wds': False,
                                'ht': True,
                                'preauth': False,
                                'assoc': True,
                                'authorized': True,
                                'vht': False,
                                'wmm': True,
                                'aid': 1,
                                'mfp': False,
                                'auth': True,
                                'signature': 'test_signature',
                            },
                            {
                                'mac': 'c0:ee:fb:34:f5:4b',
                                'wps': False,
                                'wds': False,
                                'ht': True,
                                'preauth': False,
                                'assoc': True,
                                'authorized': True,
                                'vht': False,
                                'wmm': True,
                                'aid': 1,
                                'mfp': False,
                                'auth': True,
                                'signature': 'test_signature',
                            },
                        ],
                    },
                },
            ],
        }

    mesh_interface = {
        "addresses": [
            {
                "address": "fe80::6670:2ff:fe3e:9e8b",
                "family": "ipv6",
                "mask": 64,
                "proto": "static",
            }
        ],
        "mac": "64:70:02:3e:9e:8b",
        "mtu": 1500,
        "multicast": True,
        "name": "mesh0-mng",
        "txqueuelen": 1000,
        "type": "wireless",
        "up": True,
        "wireless": {
            "channel": 11,
            "clients": [
                {
                    "auth": True,
                    "authorized": True,
                    "ht": True,
                    "mac": "A0:F3:C1:A5:FA:35",
                    "mfp": False,
                    "noise": -95,
                    "signal": 4,
                    "vht": False,
                    "wmm": True,
                }
            ],
            "country": "ES",
            "frequency": 2462,
            "htmode": "HT20",
            "mode": "802.11s",
            "noise": -95,
            "signal": 4,
            "ssid": "battlemesh-mng",
            "tx_power": 17,
        },
    }


class DeviceMonitoringTestCase(TestDeviceMonitoringMixin, TestCase):
    pass


class DeviceMonitoringTransactionTestcase(
    TestDeviceMonitoringMixin, TransactionTestCase
):
    pass


class TestWifiClientSessionMixin(TestDeviceMonitoringMixin):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        manage_short_retention_policy()

    @property
    def _sample_data(self):
        data = deepcopy(self._data())
        data.pop('resources')
        return data

    def _create_device_data(self, device=None):
        device = device or self._create_device()
        return DeviceData(pk=device.pk)

    def _save_device_data(self, device_data=None, data=None):
        dd = device_data or self._create_device_data()
        dd.data = data or self._sample_data
        dd.save_data()
        return dd

    def _create_wifi_client(self, **kwargs):
        options = {
            'mac_address': '22:33:44:55:66:77',
            'vendor': '',
            'ht': True,
            'vht': True,
            'wmm': False,
            'wds': False,
            'wps': False,
        }
        options.update(**kwargs)
        wifi_client = WifiClient(**options)
        wifi_client.full_clean()
        wifi_client.save()
        return wifi_client

    def _create_wifi_session(self, **kwargs):
        if 'wifi_client' not in kwargs:
            kwargs['wifi_client'] = self._create_wifi_client()
        if 'device' not in kwargs:
            kwargs['device'] = self._create_device()
        options = {'ssid': 'Free Public WiFi', 'interface_name': 'wlan0'}
        options.update(kwargs)
        wifi_session = WifiSession(**options)
        wifi_session.full_clean()
        wifi_session.save()
        return wifi_session



from django.db import migrations


def populate_is_healthy_tolerant(apps, schema_editor):
    Metric = apps.get_model('monitoring', 'Metric')
    chunk_size = 2000
    metrics = []
    for metric in (
        Metric.objects.filter(is_healthy__isnull=False)
        .only('id', 'is_healthy')
        .iterator(chunk_size=chunk_size)
    ):
        metric.is_healthy_tolerant = metric.is_healthy
        metrics.append(metric)
        if len(metrics) == chunk_size:
            Metric.objects.bulk_update(metrics, fields=['is_healthy_tolerant'])
            metrics = []
    if metrics:
        Metric.objects.bulk_update(metrics, fields=['is_healthy_tolerant'])


class Migration(migrations.Migration):
    dependencies = [
        ('monitoring', '0002_metric_is_healthy_tolerance'),
    ]

    operations = [
        migrations.RunPython(
            populate_is_healthy_tolerant, reverse_code=migrations.RunPython.noop
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from swapper import load_model

Check = load_model('check', 'Check')
Metric = load_model('monitoring', 'Metric')
Device = load_model('config', 'Device')


class BaseCheck(object):
    def __init__(self, check, params):
        self.check_instance = check
        self.related_object = check.content_object
        self.params = params

    def validate_instance(self):
        # check instance is of type device
        obj = self.related_object
        if not obj or not isinstance(obj, Device):
            message = 'A related device is required to perform this operation'
            raise ValidationError({'content_type': message, 'object_id': message})

    def validate(self):
        self.validate_instance()
        self.validate_params()

    def validate_params(self):
        pass

    def check(self, store=True):
        raise NotImplementedError

    def _get_or_create_metric(self, configuration=None):
        """Gets or creates metric."""
        check = self.check_instance
        if check.object_id and check.content_type_id:
            obj_id = check.object_id
            ct = ContentType.objects.get_for_id(check.content_type_id)
        else:
            obj_id = str(check.id)
            ct = ContentType.objects.get_for_model(Check)
        options = dict(
            object_id=obj_id,
            content_type_id=ct.id,
            configuration=configuration or self.__class__.__name__.lower(),
        )
        metric, created = Metric._get_or_create(**options)
        return metric, created



from django.core.management.base import BaseCommand

from ...utils import run_checks_async


class BaseRunChecksCommand(BaseCommand):
    help = 'Run all monitoring checks asynchronously'

    def handle(self, *args, **options):
        run_checks_async()



import subprocess

from django.core.exceptions import ValidationError
from jsonschema import draft7_format_checker, validate
from jsonschema.exceptions import ValidationError as SchemaError
from swapper import load_model

from openwisp_utils.utils import deep_merge_dicts

from ... import settings as monitoring_settings
from .. import settings as app_settings
from ..exceptions import OperationalError
from .base import BaseCheck

Chart = load_model('monitoring', 'Chart')
Metric = load_model('monitoring', 'Metric')
AlertSettings = load_model('monitoring', 'AlertSettings')

DEFAULT_PING_CHECK_CONFIG = {
    'count': {
        'type': 'integer',
        'default': 5,
        'minimum': 2,
        # chosen to avoid slowing down the queue
        'maximum': 20,
    },
    'interval': {
        'type': 'integer',
        'default': 25,
        'minimum': 10,
        # chosen to avoid slowing down the queue
        'maximum': 1000,
    },
    'bytes': {'type': 'integer', 'default': 56, 'minimum': 12, 'maximum': 65508},
    'timeout': {
        'type': 'integer',
        'default': 800,
        'minimum': 5,
        # arbitrary chosen to avoid slowing down the queue
        'maximum': 1500,
    },
}


def get_ping_schema():
    schema = {
        '$schema': 'http://json-schema.org/draft-07/schema#',
        'type': 'object',
        'additionalProperties': False,
    }
    schema['properties'] = deep_merge_dicts(
        DEFAULT_PING_CHECK_CONFIG, app_settings.PING_CHECK_CONFIG
    )
    return schema


class Ping(BaseCheck):
    schema = get_ping_schema()

    def validate_params(self):
        try:
            validate(self.params, self.schema, format_checker=draft7_format_checker)
        except SchemaError as e:
            message = 'Invalid param'
            path = '/'.join(e.path)
            if path:
                message = '{0} in "{1}"'.format(message, path)
            message = '{0}: {1}'.format(message, e.message)
            raise ValidationError({'params': message}) from e

    def check(self, store=True):
        count = self._get_param('count')
        interval = self._get_param('interval')
        bytes_ = self._get_param('bytes')
        timeout = self._get_param('timeout')
        ip = self._get_ip()
        #  if the device has no available IP
        if not ip:
            monitoring = self.related_object.monitoring
            # device not known yet, ignore
            if monitoring.status == 'unknown':
                return
            # device is known, simulate down
            result = {'reachable': 0, 'loss': 100.0}
            if store:
                self.store_result(result)
            return result
        command = [
            'fping',
            '-e',  # show elapsed (round-trip) time of packets
            '-c %s' % count,  # count of pings to send to each target,
            '-p %s' % interval,  # interval between sending pings(in ms)
            '-b %s' % bytes_,  # amount of ping data to send
            '-t %s' % timeout,  # individual target initial timeout (in ms)
            '-q',
            ip,
        ]
        stdout, stderr = self._command(command)
        # fpings shows statistics on stderr
        output = stderr.decode('utf8')
        try:
            parts = output.split('=')
            if len(parts) > 2:
                min, avg, max = parts[-1].strip().split('/')
                i = -2
            else:
                i = -1
            sent, received, loss = parts[i].strip().split(',')[0].split('/')
            loss = float(loss.strip('%'))
        except (IndexError, ValueError) as e:
            message = 'Unrecognized fping output:\n\n{0}'.format(output)
            raise OperationalError(message) from e
        result = {'reachable': int(loss < 100), 'loss': loss}
        if result['reachable']:
            result.update(
                {'rtt_min': float(min), 'rtt_avg': float(avg), 'rtt_max': float(max)}
            )
        if store:
            self.store_result(result)
        return result

    def store_result(self, result):
        """Stores result in the DB."""
        metric = self._get_metric()
        copied = result.copy()
        reachable = copied.pop('reachable')
        metric.write(reachable, extra_values=copied)

    def _get_param(self, param):
        """Gets specified param or its default value according to the schema."""
        return self.params.get(param, self.schema['properties'][param]['default'])

    def _get_ip(self):
        """Figures out ip to use or fails raising OperationalError."""
        device = self.related_object
        ip = device.management_ip
        if not ip and not app_settings.MANAGEMENT_IP_ONLY:
            ip = device.last_ip
        return ip

    def _command(self, command):
        """Executes command (easier to mock)."""
        p = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return p.stdout, p.stderr

    def _get_metric(self):
        """Gets or creates metric."""
        metric, created = self._get_or_create_metric()
        if created:
            self._create_alert_settings(metric)
            self._create_charts(metric)
        return metric

    def _create_alert_settings(self, metric):
        alert_settings = AlertSettings(metric=metric)
        alert_settings.full_clean()
        alert_settings.save()

    def _create_charts(self, metric):
        """Creates device charts if necessary."""
        charts = ['uptime', 'packet_loss', 'rtt']
        for chart in charts:
            if chart not in monitoring_settings.AUTO_CHARTS:
                continue
            chart = Chart(metric=metric, configuration=chart)
            chart.full_clean()
            chart.save()



from unittest.mock import patch

from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from django.utils.translation import gettext_lazy as _

from .. import settings as app_settings
from ..configuration import (
    DEFAULT_METRICS,
    get_metric_configuration,
    get_metric_configuration_choices,
    register_metric,
    unregister_metric,
)
from . import TestMonitoringMixin


class TestConfiguration(TestMonitoringMixin, TestCase):
    def _get_new_metric(self):
        return {
            'label': _('Histogram'),
            'name': 'Histogram',
            'key': 'histogram',
            'field_name': 'value',
            'charts': {
                'histogram': {
                    'type': 'histogram',
                    'title': 'Histogram',
                    'description': 'Histogram',
                    'top_fields': 2,
                    'order': 999,
                    'query': {
                        'influxdb': (
                            "SELECT {fields|SUM|/ 1} FROM {key} "
                            "WHERE time >= '{time}' AND content_type = "
                            "'{content_type}' AND object_id = '{object_id}'"
                        )
                    },
                }
            },
        }

    def test_register_invalid_metric_config(self):
        with self.subTest('Test Invalid Metric name'):
            with self.assertRaisesMessage(
                ImproperlyConfigured, 'Metric configuration name should be type "str".'
            ):
                register_metric(list(), {})
        with self.subTest('Test Invalid Metric Configuration'):
            with self.assertRaisesMessage(
                ImproperlyConfigured, 'Metric configuration should be type "dict".'
            ):
                register_metric('dummy_test', list())

    def test_unregister_invalid_metric_config(self):
        with self.subTest('Test invalid metric name'):
            with self.assertRaisesMessage(
                ImproperlyConfigured, 'Metric configuration name should be type "str".'
            ):
                unregister_metric(list())
        with self.subTest('Test unregister unregistered metric configuration'):
            with self.assertRaisesMessage(
                ImproperlyConfigured, 'No such Chart configuation "invalid".'
            ):
                unregister_metric('invalid')

    def test_register_duplicate_metric(self):
        m = self._create_general_metric()
        metric_config = m.config_dict
        with self.assertRaisesMessage(
            ImproperlyConfigured,
            f'{m.configuration} is an already registered Metric Configuration.',
        ):
            register_metric(m.configuration, metric_config)

    @patch.dict(DEFAULT_METRICS)
    def test_register_metric_configuration(self):
        metric_config = self._get_new_metric()
        register_metric('histogram', metric_config)
        self.assertIn(metric_config, get_metric_configuration().values())
        self.assertIn(('histogram', 'Histogram'), get_metric_configuration_choices())
        unregister_metric('histogram')

    @patch.dict(DEFAULT_METRICS)
    @patch.object(
        app_settings,
        'ADDITIONAL_METRICS',
        {
            'histogram': {
                'partial': True,
                'name': 'Partial Updated',
                'charts': {
                    'histogram': {
                        'title': 'Partial Updated',
                    }
                },
            }
        },
    )
    def test_overriding_registered_metric_configuration(self):
        metric_config = self._get_new_metric()
        register_metric('histogram', metric_config)
        metrics = get_metric_configuration()
        self.assertIn('histogram', metrics.keys())
        self.assertEqual(metrics['histogram']['name'], 'Partial Updated')
        self.assertEqual(
            metrics['histogram']['charts']['histogram']['title'], 'Partial Updated'
        )



from datetime import timedelta
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.test import TestCase
from django.utils import timezone
from freezegun import freeze_time
from swapper import load_model

from openwisp_utils.tests import catch_signal

from ..exceptions import InvalidChartConfigException, InvalidMetricConfigException
from ..signals import post_metric_write, pre_metric_write, threshold_crossed
from . import TestMonitoringMixin

start_time = timezone.now()
ten_minutes_ago = start_time - timedelta(minutes=10)
Metric = load_model('monitoring', 'Metric')
AlertSettings = load_model('monitoring', 'AlertSettings')
Notification = load_model('openwisp_notifications', 'Notification')


class TestModels(TestMonitoringMixin, TestCase):
    def tearDown(self):
        cache.clear()
        super().tearDown()

    def test_general_metric_str(self):
        m = Metric(name='Test metric')
        self.assertEqual(str(m), m.name)

    def test_chart_str(self):
        c = self._create_chart()
        self.assertEqual(str(c), c.label)

    def test_chart_no_valid_config(self):
        c = self._create_chart()
        c.configuration = 'invalid'
        try:
            c.full_clean()
        except ValidationError as e:
            self.assertIn('configuration', e.message_dict)
        else:
            self.fail()

    def test_metric_no_valid_config(self):
        m = self._create_object_metric()
        m.configuration = 'invalid'
        with self.assertRaises(InvalidMetricConfigException):
            m.full_clean()

    def test_invalid_chart_config(self):
        c = self._create_chart(test_data=False)
        c.configuration = 'invalid'
        with self.assertRaises(InvalidChartConfigException):
            c.config_dict

    def test_metric_related_fields(self):
        m = self._create_object_metric()
        self.assertEqual(m.related_fields, [])
        m.configuration = 'ping'
        m.full_clean()
        self.assertEqual(m.related_fields, ['loss', 'rtt_min', 'rtt_max', 'rtt_avg'])

    def test_general_codename(self):
        m = Metric(name='Test metric-(1)')
        self.assertEqual(m.codename, 'test_metric_1')
        m = Metric()
        self.assertEqual(m.codename, '')

    def test_metric_key_contains_dash(self):
        m = self._create_general_metric(key='br-lan')
        self.assertEqual(m.key, 'br_lan')

    def test_metric_key_contains_dot(self):
        m = self._create_general_metric(key='eth0.2')
        self.assertEqual(m.key, 'eth0_2')

    def test_object_metric_str(self):
        obj = self._create_user()
        om = self._create_object_metric(name='Logins', content_object=obj)
        om.refresh_from_db()
        expected = '{0} (User: {1})'.format(om.name, obj)
        self.assertEqual(str(om), expected)

    def test_general_key(self):
        m = self._create_general_metric()
        self.assertEqual(m.key, m.codename)

    def test_object_key(self):
        om = self._create_object_metric()
        self.assertEqual(om.key, om.codename)

    def test_custom_get_or_create(self):
        m, created = Metric._get_or_create(name='lan', configuration='test_metric')
        self.assertTrue(created)
        m2, created = Metric._get_or_create(name='lan', configuration='test_metric')
        self.assertEqual(m.id, m2.id)
        self.assertFalse(created)

    def test_get_or_create_renamed(self):
        m, created = Metric._get_or_create(name='lan', configuration='test_metric')
        self.assertTrue(created)
        m.name = 'renamed'
        m.save()
        m2, created = Metric._get_or_create(name='lan', configuration='test_metric')
        self.assertEqual(m.id, m2.id)
        self.assertEqual(m2.name, m.name)
        self.assertFalse(created)

    def test_get_or_create_renamed_object(self):
        obj = self._create_user()
        ct = ContentType.objects.get_for_model(get_user_model())
        m, created = Metric._get_or_create(
            name='logins',
            configuration='test_metric',
            content_type_id=ct.id,
            object_id=obj.pk,
        )
        self.assertTrue(created)
        m.name = 'renamed'
        m.save()
        m2, created = Metric._get_or_create(
            name='logins',
            configuration='test_metric',
            content_type_id=ct.id,
            object_id=obj.pk,
        )
        self.assertEqual(m.id, m2.id)
        self.assertEqual(m2.name, m.name)
        self.assertFalse(created)

    def test_get_or_create_integrity_error(self):
        with patch.object(
            Metric.objects,
            'get',
            side_effect=[
                Metric.DoesNotExist,
                Metric(name='lan', configuration='test_metric'),
            ],
        ) as mocked_get, patch.object(
            Metric, 'save', side_effect=IntegrityError
        ) as mocked_save:
            metric, _ = Metric._get_or_create(name='lan', configuration='test_metric')
            mocked_save.assert_called_once()
            self.assertEqual(mocked_get.call_count, 2)
            self.assertEqual(metric.name, 'lan')
            self.assertEqual(metric.configuration, 'test_metric')

    def test_metric_write_wrong_related_fields(self):
        m = self._create_general_metric(name='ping', configuration='ping')
        extra_values = {'reachable': 0, 'rtt_avg': 0.51, 'rtt_max': 0.6, 'rtt_min': 0.4}
        with self.assertRaisesRegex(
            ValueError, '"reachable" not defined in metric configuration'
        ):
            m.write(1, extra_values=extra_values)

    def test_batch_metric_write_wrong_related_fields(self):
        m = self._create_general_metric(name='ping', configuration='ping')
        extra_values = {'reachable': 0, 'rtt_avg': 0.51, 'rtt_max': 0.6, 'rtt_min': 0.4}
        with self.assertRaises(ValueError) as error:
            Metric.batch_write(
                [
                    (
                        m,
                        {
                            'value': 1,
                            'extra_values': extra_values,
                        },
                    ),
                ]
            )
        self.assertEqual(
            error.exception.args[0]['ping'],
            '"reachable" not defined in metric configuration',
        )

    def test_batch_write_historical_data(self):
        m = self._create_general_metric(name="load")
        self._create_alert_settings(
            metric=m, custom_operator=">", custom_threshold=90, custom_tolerance=0
        )
        with catch_signal(threshold_crossed) as handler:
            Metric.batch_write(
                [
                    (
                        m,
                        {
                            "value": 91,
                            "current": False,
                            "time": start_time - timedelta(minutes=5),
                        },
                    )
                ]
            )
        handler.assert_not_called()

    def test_tags(self):
        extra_tags = {'a': 'a', 'b': 'b1'}
        metric = self._create_object_metric(extra_tags=extra_tags)
        expected_tags = extra_tags.copy()
        expected_tags.update(
            {'object_id': metric.object_id, 'content_type': metric.content_type_key}
        )
        self.assertEqual(metric.tags, expected_tags)

    def test_read_general_metric(self):
        m = self._create_general_metric(name='load')
        m.write(50, check=False)
        self.assertEqual(self._read_metric(m)[0][m.field_name], 50)
        m.write(1, check=False)
        self.assertEqual(self._read_metric(m)[0][m.field_name], 50)
        self.assertEqual(self._read_metric(m, order='-time')[0][m.field_name], 1)

    def test_read_object_metric(self):
        om = self._create_object_metric(name='load')
        om.write(50)
        om.write(3)
        self._read_metric(om, extra_fields='*')
        self.assertEqual(self._read_metric(om)[0][om.field_name], 50)

    def test_alert_settings_max_seconds(self):
        m = self._create_general_metric(name='load')
        try:
            self._create_alert_settings(
                metric=m,
                custom_operator='>',
                custom_threshold=90,
                custom_tolerance=9999999,
            )
        except ValidationError as e:
            self.assertIn('custom_tolerance', e.message_dict)
        else:
            self.fail('ValidationError not raised')

    def test_threshold_is_crossed_error(self):
        m = self._create_general_metric(name='load')
        alert_s = self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=0
        )
        with self.assertRaises(ValueError):
            alert_s._is_crossed_by(alert_s, start_time)

    def test_threshold_is_crossed_immediate(self):
        m = self._create_general_metric(name='load')
        alert_s = self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=0
        )
        self.assertFalse(alert_s._is_crossed_by(80, start_time))
        self.assertTrue(alert_s._is_crossed_by(91, start_time))
        self.assertTrue(alert_s._is_crossed_by(100, start_time))
        self.assertTrue(alert_s._is_crossed_by(100))
        self.assertFalse(alert_s._is_crossed_by(90, start_time))
        alert_s.custom_operator = '<'
        alert_s.save()
        self.assertTrue(alert_s._is_crossed_by(80))

    def test_threshold_is_crossed_deferred(self):
        m = self._create_general_metric(name='load')
        alert_s = self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=9
        )
        self.assertFalse(alert_s._is_crossed_by(95, start_time))
        self.assertTrue(alert_s._is_crossed_by(95, ten_minutes_ago))
        self.assertFalse(alert_s._is_crossed_by(80, start_time))
        self.assertFalse(alert_s._is_crossed_by(80, ten_minutes_ago))

    def test_threshold_is_crossed_deferred_2(self):
        self._create_admin()
        m = self._create_general_metric(name='load')
        self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=1
        )
        m.write(60)
        m.write(99)
        m.refresh_from_db(fields=['is_healthy', 'is_healthy_tolerant'])
        self.assertEqual(m.is_healthy, False)
        self.assertEqual(m.is_healthy_tolerant, True)

    def test_general_check_threshold_no_exception(self):
        m = self._create_general_metric()
        m.check_threshold(1)

    def test_general_metric_signal_emitted(self):
        m = self._create_general_metric(name='load')
        alert_s = self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=0
        )
        with catch_signal(threshold_crossed) as handler:
            m.check_threshold(91)
        handler.assert_called_once_with(
            alert_settings=alert_s,
            first_time=False,
            metric=m,
            target=None,
            sender=Metric,
            signal=threshold_crossed,
            tolerance_crossed=True,
        )

    def test_object_metric_signal_emitted(self):
        om = self._create_object_metric()
        alert_s = self._create_alert_settings(
            metric=om, custom_operator='>', custom_threshold=90, custom_tolerance=0
        )
        with catch_signal(threshold_crossed) as handler:
            om.check_threshold(91)
        handler.assert_called_once_with(
            alert_settings=alert_s,
            first_time=False,
            metric=om,
            target=om.content_object,
            sender=Metric,
            signal=threshold_crossed,
            tolerance_crossed=True,
        )

    def test_metric_pre_write_signals_emitted(self):
        om = self._create_object_metric()
        with catch_signal(pre_metric_write) as handler:
            om.write(3)
            handler.assert_called_once_with(
                sender=Metric,
                metric=om,
                values={om.field_name: 3},
                signal=pre_metric_write,
                time=None,
                current=False,
            )

    def test_metric_post_write_signals_emitted(self):
        om = self._create_object_metric()
        with catch_signal(post_metric_write) as handler:
            om.write(3, current=True, time=start_time)
            handler.assert_called_once_with(
                sender=Metric,
                metric=om,
                values={om.field_name: 3},
                signal=post_metric_write,
                time=start_time.isoformat(),
                current=True,
            )

    def test_clean_default_threshold_values(self):
        m = self._create_general_metric(configuration='ping')
        alert_s = AlertSettings(metric=m)
        with self.subTest('Store default values'):
            alert_s.custom_operator = alert_s.operator
            alert_s.custom_threshold = alert_s.threshold
            alert_s.custom_tolerance = alert_s.tolerance
            alert_s.full_clean()
            self.assertIsNone(alert_s.custom_operator)
            self.assertIsNone(alert_s.custom_threshold)
            self.assertIsNone(alert_s.custom_tolerance)
        with self.subTest('Store default and custom values'):
            alert_s.custom_operator = alert_s.operator
            alert_s.custom_threshold = 0.5
            alert_s.custom_tolerance = 2
            alert_s.full_clean()
            self.assertIsNone(alert_s.custom_operator)
            self.assertEqual(alert_s.custom_threshold, 0.5)
            self.assertEqual(alert_s.custom_tolerance, 2)

    def test_alert_settings_tolerance_default(self):
        m = self._create_general_metric(name='load')
        alert_s = AlertSettings(metric=m)
        self.assertIsNone(alert_s.custom_tolerance)

    def test_tolerance(self):
        self._create_admin()
        m = self._create_general_metric(name='load')
        self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=5
        )
        with freeze_time(start_time):
            m.write(99)
        with self.subTest("within tolerance, no alerts expected"):
            with freeze_time(start_time + timedelta(minutes=2)):
                m.write(99)
            m.refresh_from_db(fields=["is_healthy", "is_healthy_tolerant"])
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 0)
            with freeze_time(start_time + timedelta(minutes=4)):
                m.write(99)
            m.refresh_from_db(fields=["is_healthy", "is_healthy_tolerant"])
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 0)
        with self.subTest("tolerance trepassed, alerts expected"):
            with freeze_time(start_time + timedelta(minutes=6)):
                m.write(99)
            m.refresh_from_db(fields=["is_healthy", "is_healthy_tolerant"])
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)
        with self.subTest("value back to normal"):
            with freeze_time(start_time + timedelta(minutes=7)):
                m.write(71)
            with freeze_time(start_time + timedelta(minutes=12)):
                m.write(71)
            m.refresh_from_db(fields=["is_healthy", "is_healthy_tolerant"])
            self.assertEqual(m.is_healthy, True)
            self.assertEqual(m.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 2)

    def test_time_crossed(self):
        m = self._create_general_metric(name='load')
        a = self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=5
        )

        now = timezone.now()
        self.assertFalse(a._time_crossed(now))
        self.assertFalse(a._time_crossed(now - timedelta(minutes=1)))
        self.assertFalse(a._time_crossed(now - timedelta(minutes=4)))
        self.assertTrue(a._time_crossed(now - timedelta(minutes=5)))
        self.assertTrue(a._time_crossed(now - timedelta(minutes=6)))

    def test_get_time_str(self):
        m = self._create_general_metric(name='load')
        now = timezone.now()
        self.assertEqual(m._get_time(now.isoformat()), now)

    def test_deleting_metric_deletes_timeseries(self):
        metric1 = self._create_general_metric(name='load')
        metric2 = self._create_general_metric(name='traffic')
        metric1.write(99)
        metric2.write(5000)
        self.assertNotEqual(self._read_metric(metric1), [])
        self.assertNotEqual(self._read_metric(metric2), [])
        metric1.delete()
        self.assertEqual(self._read_metric(metric1), [])
        # Only the timeseries data related to the deleted metric
        # should be deleted
        self.assertNotEqual(self._read_metric(metric2), [])

    def test_metric_invalid_field_name(self):
        metric = self._create_general_metric(configuration='test_alert_field')
        metric.field_name = 'invalid_field'
        with self.assertRaises(ValidationError) as err:
            metric.full_clean()
            metric.save()
        self.assertIn(
            f'"{metric.field_name}" must be one of the following metric fields',
            str(err.exception),
        )

    def test_alert_field_property(self):
        m = self._create_general_metric(configuration='test_alert_field')
        # When metric field_name == config['field_name']
        self.assertEqual(m.field_name, 'test_alert_field')
        # Get the alert_field from the config
        self.assertEqual(m.alert_field, 'test_related_2')
        m.field_name = 'test_related_3'
        m.full_clean()
        m.save()
        # When metric field_name != config['field_name']
        self.assertEqual(m.field_name, 'test_related_3')
        # alert_field same as field_name
        self.assertEqual(m.alert_field, 'test_related_3')



from django.dispatch import Signal

health_status_changed = Signal()
health_status_changed.__doc__ = """
Providing arguments: ['instance', 'status']
"""
device_metrics_received = Signal()
device_metrics_received.__doc__ = """
Providing arguments: ['instance', 'request', 'time', 'current']
"""



from ..db import timeseries_db
from . import settings as app_settings

SHORT_RP = 'short'
DEFAULT_RP = 'autogen'


def get_device_cache_key(device, context='react-to-updates'):
    return f'device-{device.pk}-{context}'


def manage_short_retention_policy():
    """creates or updates the "short" retention policy"""
    duration = app_settings.SHORT_RETENTION_POLICY
    timeseries_db.create_or_alter_retention_policy(SHORT_RP, duration)


def manage_default_retention_policy():
    """creates or updates the "default" retention policy"""
    duration = app_settings.DEFAULT_RETENTION_POLICY
    timeseries_db.create_or_alter_retention_policy(DEFAULT_RP, duration)



from json import loads
from unittest.mock import call, patch

from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.test import TransactionTestCase
from swapper import load_model

from openwisp_controller.connection.connectors.ssh import Ssh
from openwisp_controller.connection.models import DeviceConnection
from openwisp_controller.connection.settings import CONNECTORS, UPDATE_STRATEGIES
from openwisp_controller.connection.tests.utils import CreateConnectionsMixin, SshServer
from openwisp_monitoring.check.classes.iperf3 import get_iperf3_schema
from openwisp_monitoring.check.classes.iperf3 import logger as iperf3_logger

from ...device.tests import TestDeviceMonitoringMixin
from .. import settings as app_settings
from ..classes import Iperf3
from .iperf3_test_utils import (
    INVALID_PARAMS,
    PARAM_ERROR,
    RESULT_AUTH_FAIL,
    RESULT_FAIL,
    RESULT_TCP,
    RESULT_UDP,
    TEST_RSA_KEY,
)

Check = load_model('check', 'Check')
Chart = load_model('monitoring', 'Chart')
Metric = load_model('monitoring', 'Metric')
AlertSettings = load_model('monitoring', 'AlertSettings')


class TestIperf3(
    CreateConnectionsMixin, TestDeviceMonitoringMixin, TransactionTestCase
):
    _IPERF3 = app_settings.CHECK_CLASSES[2][0]
    _RESULT_KEYS = [
        'iperf3_result',
        'sent_bps_tcp',
        'received_bps_tcp',
        'sent_bytes_tcp',
        'received_bytes_tcp',
        'retransmits',
        'sent_bps_udp',
        'sent_bytes_udp',
        'jitter',
        'total_packets',
        'lost_packets',
        'lost_percent',
    ]
    _IPERF3_TEST_SERVER = ['iperf3.openwisptestserver.com']
    _IPERF3_TEST_MULTIPLE_SERVERS = [
        'iperf3.openwisptestserver1.com',
        'iperf3.openwisptestserver2.com',
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.mock_ssh_server = SshServer(
            {'root': cls._TEST_RSA_PRIVATE_KEY_PATH}
        ).__enter__()
        cls.ssh_server.port = cls.mock_ssh_server.port

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.mock_ssh_server.__exit__()
        app_settings.IPERF3_CHECK_CONFIG = {}

    def setUp(self):
        ckey = self._create_credentials_with_key(port=self.ssh_server.port)
        self.dc = self._create_device_connection(credentials=ckey)
        self.device = self.dc.device
        self.org_id = str(self.device.organization.id)
        self.dc.connect()
        app_settings.IPERF3_CHECK_CONFIG = {
            self.org_id: {'host': self._IPERF3_TEST_SERVER}
        }
        self._EXPECTED_COMMAND_CALLS = [
            call(
                (
                    'iperf3 -c iperf3.openwisptestserver.com -p 5201 -t 10 --connect-timeout 1000 '
                    '-b 0 -l 128K -w 0 -P 1  -J'
                ),
                raise_unexpected_exit=False,
            ),
            call(
                (
                    'iperf3 -c iperf3.openwisptestserver.com -p 5201 -t 10 --connect-timeout 1000 '
                    '-b 30M -l 0 -w 0 -P 1  -u -J'
                ),
                raise_unexpected_exit=False,
            ),
        ]
        self._EXPECTED_WARN_CALLS = [
            call(
                (
                    f'Iperf3 check failed for "{self.device}", '
                    'error - unable to connect to server: Connection refused'
                )
            ),
            call(
                (
                    f'Iperf3 check failed for "{self.device}", '
                    'error - unable to connect to server: Connection refused'
                )
            ),
        ]

    def _perform_iperf3_check(self, object_id=None):
        if not object_id:
            object_id = self.device.id
        check = Check.objects.get(check_type=self._IPERF3, object_id=object_id)
        return check.perform_check(store=False)

    def _set_auth_expected_calls(self, config):
        password = config[self.org_id]['password']
        username = config[self.org_id]['username']
        server = 'iperf3.openwisptestserver.com'
        test_prefix = '-----BEGIN PUBLIC KEY-----\n'
        test_suffix = '\n-----END PUBLIC KEY-----'
        key = config[self.org_id]['rsa_public_key']
        rsa_key_path = '/tmp/iperf3-public-key.pem'

        self._EXPECTED_COMMAND_CALLS = [
            call(
                (
                    f'echo "{test_prefix}{key}{test_suffix}" > {rsa_key_path} && '
                    f'IPERF3_PASSWORD="{password}" iperf3 -c {server} -p 5201 -t 10 '
                    f'--username "{username}" --rsa-public-key-path {rsa_key_path} --connect-timeout 1000 '
                    f'-b 0 -l 128K -w 0 -P 1  -J'
                ),
                raise_unexpected_exit=False,
            ),
            call(
                (
                    f'IPERF3_PASSWORD="{password}" iperf3 -c {server} -p 5201 -t 10 '
                    f'--username "{username}" --rsa-public-key-path {rsa_key_path} --connect-timeout 1000 '
                    f'-b 30M -l 0 -w 0 -P 1  -u -J '
                    f'&& rm -f {rsa_key_path}'
                ),
                raise_unexpected_exit=False,
            ),
        ]

    def _assert_iperf3_fail_result(self, result):
        for key in self._RESULT_KEYS:
            self.assertIn(key, result)
        self.assertEqual(result['iperf3_result'], 0)
        self.assertEqual(result['sent_bps_tcp'], 0.0)
        self.assertEqual(result['received_bps_tcp'], 0.0)
        self.assertEqual(result['sent_bytes_tcp'], 0)
        self.assertEqual(result['received_bytes_tcp'], 0)
        self.assertEqual(result['retransmits'], 0)
        self.assertEqual(result['sent_bps_udp'], 0.0)
        self.assertEqual(result['sent_bytes_udp'], 0)
        self.assertEqual(result['jitter'], 0.0)
        self.assertEqual(result['total_packets'], 0)
        self.assertEqual(result['lost_percent'], 0.0)

    @patch.object(Ssh, 'exec_command')
    @patch.object(iperf3_logger, 'warning')
    def test_iperf3_check_no_params(self, mock_warn, mock_exec_command):
        mock_exec_command.side_effect = [(RESULT_TCP, 0), (RESULT_UDP, 0)]
        # By default check params {}
        tcp_result = loads(RESULT_TCP)['end']
        udp_result = loads(RESULT_UDP)['end']['sum']
        result = self._perform_iperf3_check()
        for key in self._RESULT_KEYS:
            self.assertIn(key, result)
        self.assertEqual(result['iperf3_result'], 1)
        self.assertEqual(
            result['sent_bps_tcp'], tcp_result['sum_sent']['bits_per_second']
        )
        self.assertEqual(
            result['received_bytes_tcp'], tcp_result['sum_received']['bytes']
        )
        self.assertEqual(result['jitter'], udp_result['jitter_ms'])
        self.assertEqual(result['total_packets'], udp_result['packets'])
        self.assertEqual(mock_warn.call_count, 0)
        self.assertEqual(mock_exec_command.call_count, 2)
        mock_exec_command.assert_has_calls(self._EXPECTED_COMMAND_CALLS)

    @patch.object(Ssh, 'exec_command')
    @patch.object(iperf3_logger, 'warning')
    def test_iperf3_check_params(self, mock_warn, mock_exec_command):
        mock_exec_command.side_effect = [(RESULT_TCP, 0), (RESULT_UDP, 0)]
        check = Check.objects.get(check_type=self._IPERF3)
        tcp_result = loads(RESULT_TCP)['end']
        udp_result = loads(RESULT_UDP)['end']['sum']
        server = self._IPERF3_TEST_SERVER[0]
        test_prefix = '-----BEGIN PUBLIC KEY-----\n'
        test_suffix = '\n-----END PUBLIC KEY-----'
        rsa_key_path = '/tmp/test-rsa.pem'
        test_params = {
            'username': 'openwisp-test-user',
            'password': 'openwisp_pass',
            'rsa_public_key': TEST_RSA_KEY,
            'client_options': {
                'port': 6201,
                'time': 20,
                'window': '300K',
                'parallel': 5,
                'reverse': True,
                'connect_timeout': 1000,
                'tcp': {'bitrate': '10M', 'length': '128K'},
                'udp': {'bitrate': '50M', 'length': '400K'},
            },
        }
        time = test_params['client_options']['time']
        port = test_params['client_options']['port']
        window = test_params['client_options']['window']
        parallel = test_params['client_options']['parallel']
        tcp_bitrate = test_params['client_options']['tcp']['bitrate']
        tcp_len = test_params['client_options']['tcp']['length']
        udp_bitrate = test_params['client_options']['udp']['bitrate']
        udp_len = test_params['client_options']['udp']['length']
        username = test_params['username']
        password = test_params['password']
        key = test_params['rsa_public_key']
        rsa_key_path = '/tmp/iperf3-public-key.pem'
        check.params = test_params
        check.save()
        self._EXPECTED_COMMAND_CALLS = [
            call(
                (
                    f'echo "{test_prefix}{key}{test_suffix}" > {rsa_key_path} && '
                    f'IPERF3_PASSWORD="{password}" iperf3 -c {server} -p {port} -t {time} '
                    f'--username "{username}" --rsa-public-key-path {rsa_key_path} --connect-timeout 1000 '
                    f'-b {tcp_bitrate} -l {tcp_len} -w {window} -P {parallel} --reverse -J'
                ),
                raise_unexpected_exit=False,
            ),
            call(
                (
                    f'IPERF3_PASSWORD="{password}" iperf3 -c {server} -p {port} -t {time} '
                    f'--username "{username}" --rsa-public-key-path {rsa_key_path} --connect-timeout 1000 '
                    f'-b {udp_bitrate} -l {udp_len} -w {window} -P {parallel} --reverse -u -J '
                    f'&& rm -f {rsa_key_path}'
                ),
                raise_unexpected_exit=False,
            ),
        ]
        result = self._perform_iperf3_check()
        for key in self._RESULT_KEYS:
            self.assertIn(key, result)
        self.assertEqual(result['iperf3_result'], 1)
        self.assertEqual(
            result['sent_bps_tcp'], tcp_result['sum_sent']['bits_per_second']
        )
        self.assertEqual(
            result['received_bytes_tcp'], tcp_result['sum_received']['bytes']
        )
        self.assertEqual(result['jitter'], udp_result['jitter_ms'])
        self.assertEqual(result['total_packets'], udp_result['packets'])
        self.assertEqual(mock_warn.call_count, 0)
        self.assertEqual(mock_exec_command.call_count, 2)
        mock_exec_command.assert_has_calls(self._EXPECTED_COMMAND_CALLS)

    @patch.object(Ssh, 'exec_command')
    @patch.object(iperf3_logger, 'warning')
    def test_iperf3_check_config(self, mock_warn, mock_exec_command):
        mock_exec_command.side_effect = [(RESULT_TCP, 0), (RESULT_UDP, 0)]
        tcp_result = loads(RESULT_TCP)['end']
        udp_result = loads(RESULT_UDP)['end']['sum']
        self._EXPECTED_COMMAND_CALLS = [
            call(
                (
                    'iperf3 -c iperf3.openwisptestserver.com -p 9201 -k 1M --connect-timeout 2000 '
                    '-b 10M -l 512K -w 0 -P 1 --bidir -J'
                ),
                raise_unexpected_exit=False,
            ),
            call(
                (
                    'iperf3 -c iperf3.openwisptestserver.com -p 9201 -k 1M --connect-timeout 2000 '
                    '-b 50M -l 256K -w 0 -P 1 --bidir -u -J'
                ),
                raise_unexpected_exit=False,
            ),
        ]
        iperf3_config = {
            self.org_id: {
                'host': ['iperf3.openwisptestserver.com'],
                'client_options': {
                    'port': 9201,
                    'time': 120,
                    'connect_timeout': 2000,
                    'bytes': '20M',
                    'blockcount': '1M',
                    'bidirectional': True,
                    'tcp': {'bitrate': '10M', 'length': '512K'},
                    'udp': {'bitrate': '50M', 'length': '256K'},
                },
            }
        }
        with patch.object(app_settings, 'IPERF3_CHECK_CONFIG', iperf3_config):
            with patch.object(Iperf3, 'schema', get_iperf3_schema()):
                result = self._perform_iperf3_check()
                for key in self._RESULT_KEYS:
                    self.assertIn(key, result)
                self.assertEqual(result['iperf3_result'], 1)
                self.assertEqual(
                    result['sent_bps_tcp'], tcp_result['sum_sent']['bits_per_second']
                )
                self.assertEqual(
                    result['received_bytes_tcp'], tcp_result['sum_received']['bytes']
                )
                self.assertEqual(result['jitter'], udp_result['jitter_ms'])
                self.assertEqual(result['total_packets'], udp_result['packets'])
                self.assertEqual(mock_warn.call_count, 0)
                self.assertEqual(mock_exec_command.call_count, 2)
                mock_exec_command.assert_has_calls(self._EXPECTED_COMMAND_CALLS)

    @patch.object(iperf3_logger, 'warning')
    def test_iperf3_device_connection(self, mock_warn):
        dc = self.dc

        with self.subTest(
            'Test iperf3 check active device connection when the management tunnel is down'
        ):
            with patch.object(
                DeviceConnection, 'connect', return_value=False
            ) as mocked_connect:
                self._perform_iperf3_check()
                mock_warn.assert_called_once_with(
                    f'Failed to get a working DeviceConnection for "{self.device}", iperf3 check skipped!'
                )
            self.assertEqual(mocked_connect.call_count, 1)
        mock_warn.reset_mock()

        with self.subTest('Test iperf3 check when device connection is not enabled'):
            dc.enabled = False
            dc.save()
            self._perform_iperf3_check()
            mock_warn.assert_called_once_with(
                f'Failed to get a working DeviceConnection for "{self.device}", iperf3 check skipped!'
            )
        mock_warn.reset_mock()

        with self.subTest(
            'Test iperf3 check with different credential connector type ie. snmp'
        ):
            device2 = self._create_device(
                name='test-device-2', mac_address='00:11:22:33:44:66'
            )
            params = {'community': 'public', 'agent': 'snmp-agent', 'port': 161}
            snmp_credentials = self._create_credentials(
                params=params, connector=CONNECTORS[1][0], auto_add=True
            )
            dc2 = self._create_device_connection(
                device=device2,
                credentials=snmp_credentials,
                update_strategy=UPDATE_STRATEGIES[0][0],
            )
            dc2.is_working = True
            dc2.enabled = True
            dc2.save()
            self._perform_iperf3_check(object_id=device2.id)
            mock_warn.assert_called_once_with(
                f'Failed to get a working DeviceConnection for "{device2}", iperf3 check skipped!'
            )

    @patch.object(iperf3_logger, 'info')
    def test_iperf3_check_device_monitoring_critical(self, mock_info):
        self.device.monitoring.update_status('critical')
        self._perform_iperf3_check()
        mock_info.assert_called_once_with(
            (
                f'"{self.device}" DeviceMonitoring '
                'health status is "critical", iperf3 check skipped!'
            )
        )

    def test_iperf3_check_content_object_none(self):
        check = Check(name='Iperf3 check', check_type=self._IPERF3, params={})
        try:
            check.check_instance.validate()
        except ValidationError as e:
            self.assertIn('device', str(e))
        else:
            self.fail('ValidationError not raised')

    def test_iperf3_check_content_object_not_device(self):
        check = Check(
            name='Iperf3 check',
            check_type=self._IPERF3,
            content_object=self._create_user(),
            params={},
        )
        try:
            check.check_instance.validate()
        except ValidationError as e:
            self.assertIn('device', str(e))
        else:
            self.fail('ValidationError not raised')

    def test_iperf3_check_schema_violation(self):
        for invalid_param in INVALID_PARAMS:
            check = Check(
                name='Iperf3 check',
                check_type=self._IPERF3,
                content_object=self.device,
                params=invalid_param,
            )
            try:
                check.check_instance.validate()
            except ValidationError as e:
                self.assertIn('Invalid param', str(e))
            else:
                self.fail('ValidationError not raised')

    @patch.object(Ssh, 'exec_command')
    @patch.object(iperf3_logger, 'warning')
    def test_iperf3_check(self, mock_warn, mock_exec_command):
        error = "ash: iperf3: not found"
        tcp_result = loads(RESULT_TCP)['end']
        udp_result = loads(RESULT_UDP)['end']['sum']
        iperf3_json_error_config = {
            self.org_id: {
                'host': ['iperf3.openwisptestserver.com'],
                'username': 'test',
                'password': 'testpass',
                'rsa_public_key': 'INVALID_RSA_KEY',
            }
        }
        with patch.object(
            app_settings, 'IPERF3_CHECK_CONFIG', iperf3_json_error_config
        ):
            with self.subTest('Test iperf3 errors not in json format'):
                mock_exec_command.side_effect = [(PARAM_ERROR, 1), (PARAM_ERROR, 1)]
                EXPECTED_WARN_CALLS = [
                    call(
                        f'Iperf3 check failed for "{self.device}", error - {PARAM_ERROR}'
                    ),
                    call(
                        f'Iperf3 check failed for "{self.device}", error - {PARAM_ERROR}'
                    ),
                ]
                self._perform_iperf3_check()
                self.assertEqual(mock_warn.call_count, 2)
                self.assertEqual(mock_exec_command.call_count, 2)
                mock_warn.assert_has_calls(EXPECTED_WARN_CALLS)
            mock_warn.reset_mock()
            mock_exec_command.reset_mock()

        with self.subTest('Test iperf3 is not installed on the device'):
            mock_exec_command.side_effect = [(error, 127)]
            self._perform_iperf3_check()
            mock_warn.assert_called_with(
                f'Iperf3 is not installed on the "{self.device}", error - {error}'
            )
            self.assertEqual(mock_warn.call_count, 1)
            self.assertEqual(mock_exec_command.call_count, 1)
        mock_warn.reset_mock()
        mock_exec_command.reset_mock()

        with self.subTest('Test iperf3 check passes in both TCP & UDP'):
            mock_exec_command.side_effect = [(RESULT_TCP, 0), (RESULT_UDP, 0)]
            self.assertEqual(Chart.objects.count(), 2)
            self.assertEqual(Metric.objects.count(), 2)
            result = self._perform_iperf3_check()
            for key in self._RESULT_KEYS:
                self.assertIn(key, result)
            self.assertEqual(result['iperf3_result'], 1)
            self.assertEqual(
                result['sent_bps_tcp'], tcp_result['sum_sent']['bits_per_second']
            )
            self.assertEqual(
                result['received_bps_tcp'],
                tcp_result['sum_received']['bits_per_second'],
            )
            self.assertEqual(result['sent_bytes_tcp'], tcp_result['sum_sent']['bytes'])
            self.assertEqual(
                result['received_bytes_tcp'], tcp_result['sum_received']['bytes']
            )
            self.assertEqual(
                result['retransmits'], tcp_result['sum_sent']['retransmits']
            )
            self.assertEqual(result['sent_bps_udp'], udp_result['bits_per_second'])
            self.assertEqual(result['sent_bytes_udp'], udp_result['bytes'])
            self.assertEqual(result['jitter'], udp_result['jitter_ms'])
            self.assertEqual(result['total_packets'], udp_result['packets'])
            self.assertEqual(result['lost_percent'], udp_result['lost_percent'])
            self.assertEqual(Chart.objects.count(), 8)
            self.assertEqual(Check.objects.count(), 3)
            iperf3_metric = Metric.objects.get(key='iperf3')
            self.assertEqual(Metric.objects.count(), 3)
            self.assertEqual(iperf3_metric.content_object, self.device)
            points = self._read_metric(
                iperf3_metric, limit=None, extra_fields=list(result.keys())
            )
            self.assertEqual(len(points), 1)
            self.assertEqual(points[0]['iperf3_result'], result['iperf3_result'])
            self.assertEqual(points[0]['sent_bps_tcp'], result['sent_bps_tcp'])
            self.assertEqual(
                points[0]['received_bytes_tcp'], result['received_bytes_tcp']
            )
            self.assertEqual(points[0]['retransmits'], result['retransmits'])
            self.assertEqual(points[0]['sent_bps_udp'], result['sent_bps_udp'])
            self.assertEqual(points[0]['sent_bytes_udp'], result['sent_bytes_udp'])
            self.assertEqual(points[0]['jitter'], result['jitter'])
            self.assertEqual(points[0]['total_packets'], result['total_packets'])
            self.assertEqual(points[0]['lost_packets'], result['lost_packets'])
            self.assertEqual(points[0]['lost_percent'], result['lost_percent'])
            self.assertEqual(mock_warn.call_count, 0)
            self.assertEqual(mock_exec_command.call_count, 2)
            mock_exec_command.assert_has_calls(self._EXPECTED_COMMAND_CALLS)
        mock_warn.reset_mock()
        mock_exec_command.reset_mock()

        with self.subTest('Test iperf3 check fails in both TCP & UDP'):
            mock_exec_command.side_effect = [(RESULT_FAIL, 1), (RESULT_FAIL, 1)]
            result = self._perform_iperf3_check()
            self._assert_iperf3_fail_result(result)
            self.assertEqual(Chart.objects.count(), 8)
            self.assertEqual(Metric.objects.count(), 3)
            self.assertEqual(mock_warn.call_count, 2)
            self.assertEqual(mock_exec_command.call_count, 2)
            mock_warn.assert_has_calls(self._EXPECTED_WARN_CALLS)
            mock_exec_command.assert_has_calls(self._EXPECTED_COMMAND_CALLS)
        mock_warn.reset_mock()
        mock_exec_command.reset_mock()

        with self.subTest('Test iperf3 check TCP pass UDP fail'):
            mock_exec_command.side_effect = [(RESULT_TCP, 0), (RESULT_FAIL, 1)]
            result = self._perform_iperf3_check()
            for key in self._RESULT_KEYS:
                self.assertIn(key, result)
            self.assertEqual(result['iperf3_result'], 1)
            self.assertEqual(
                result['sent_bps_tcp'], tcp_result['sum_sent']['bits_per_second']
            )
            self.assertEqual(
                result['received_bps_tcp'],
                tcp_result['sum_received']['bits_per_second'],
            )
            self.assertEqual(result['sent_bytes_tcp'], tcp_result['sum_sent']['bytes'])
            self.assertEqual(
                result['received_bytes_tcp'], tcp_result['sum_received']['bytes']
            )
            self.assertEqual(
                result['retransmits'], tcp_result['sum_sent']['retransmits']
            )
            self.assertEqual(result['sent_bps_udp'], 0.0)
            self.assertEqual(result['sent_bytes_udp'], 0)
            self.assertEqual(result['jitter'], 0.0)
            self.assertEqual(result['total_packets'], 0)
            self.assertEqual(result['lost_percent'], 0.0)
            self.assertEqual(Chart.objects.count(), 8)
            self.assertEqual(Metric.objects.count(), 3)
            self.assertEqual(mock_warn.call_count, 1)
            self.assertEqual(mock_exec_command.call_count, 2)
            mock_warn.assert_has_calls(self._EXPECTED_WARN_CALLS[1:])
            mock_exec_command.assert_has_calls(self._EXPECTED_COMMAND_CALLS)
        mock_warn.reset_mock()
        mock_exec_command.reset_mock()

        with self.subTest('Test iperf3 check TCP fail UDP pass'):
            mock_exec_command.side_effect = [(RESULT_FAIL, 1), (RESULT_UDP, 0)]
            result = self._perform_iperf3_check()
            for key in self._RESULT_KEYS:
                self.assertIn(key, result)
            self.assertEqual(result['iperf3_result'], 1)
            self.assertEqual(result['sent_bps_tcp'], 0.0)
            self.assertEqual(result['received_bps_tcp'], 0.0)
            self.assertEqual(result['sent_bytes_tcp'], 0)
            self.assertEqual(result['received_bytes_tcp'], 0)
            self.assertEqual(result['retransmits'], 0)
            self.assertEqual(result['sent_bps_udp'], udp_result['bits_per_second'])
            self.assertEqual(result['sent_bytes_udp'], udp_result['bytes'])
            self.assertEqual(result['jitter'], udp_result['jitter_ms'])
            self.assertEqual(result['total_packets'], udp_result['packets'])
            self.assertEqual(result['lost_percent'], udp_result['lost_percent'])
            self.assertEqual(Chart.objects.count(), 8)
            self.assertEqual(Metric.objects.count(), 3)
            self.assertEqual(mock_warn.call_count, 1)
            self.assertEqual(mock_exec_command.call_count, 2)
            mock_warn.assert_has_calls(self._EXPECTED_WARN_CALLS[1:])
            mock_exec_command.assert_has_calls(self._EXPECTED_COMMAND_CALLS)

    @patch.object(Ssh, 'exec_command')
    @patch.object(iperf3_logger, 'warning')
    def test_iperf3_check_auth_config(self, mock_warn, mock_exec_command):
        iperf3_config = {
            self.org_id: {
                'host': self._IPERF3_TEST_SERVER,
                'username': 'test',
                'password': 'testpass',
                'rsa_public_key': TEST_RSA_KEY,
            }
        }
        iperf3_conf_wrong_pass = {
            self.org_id: {
                'host': self._IPERF3_TEST_SERVER,
                'username': 'test',
                'password': 'wrongpass',
                'rsa_public_key': TEST_RSA_KEY,
            }
        }
        iperf3_conf_wrong_user = {
            self.org_id: {
                'host': self._IPERF3_TEST_SERVER,
                'username': 'wronguser',
                'password': 'testpass',
                'rsa_public_key': TEST_RSA_KEY,
            }
        }
        auth_error = "test authorization failed"
        tcp_result = loads(RESULT_TCP)['end']
        udp_result = loads(RESULT_UDP)['end']['sum']

        self._EXPECTED_WARN_CALLS = [
            call(f'Iperf3 check failed for "{self.device}", error - {auth_error}'),
            call(f'Iperf3 check failed for "{self.device}", error - {auth_error}'),
        ]
        with self.subTest('Test iperf3 check with right config'):
            with patch.object(
                app_settings,
                'IPERF3_CHECK_CONFIG',
                iperf3_config,
                # It is required to mock "Iperf3.schema" here so that it
                # uses the updated configuration from "IPERF3_CHECK_CONFIG" setting.
            ), patch.object(Iperf3, 'schema', get_iperf3_schema()):
                self._set_auth_expected_calls(iperf3_config)
                mock_exec_command.side_effect = [(RESULT_TCP, 0), (RESULT_UDP, 0)]
                result = self._perform_iperf3_check()
                for key in self._RESULT_KEYS:
                    self.assertIn(key, result)
                self.assertEqual(result['iperf3_result'], 1)
                self.assertEqual(
                    result['sent_bps_tcp'], tcp_result['sum_sent']['bits_per_second']
                )
                self.assertEqual(
                    result['received_bytes_tcp'], tcp_result['sum_received']['bytes']
                )
                self.assertEqual(result['jitter'], udp_result['jitter_ms'])
                self.assertEqual(result['total_packets'], udp_result['packets'])
                self.assertEqual(mock_warn.call_count, 0)
                self.assertEqual(mock_exec_command.call_count, 2)
                mock_exec_command.assert_has_calls(self._EXPECTED_COMMAND_CALLS)
            mock_warn.reset_mock()
            mock_exec_command.reset_mock()

        with self.subTest('Test iperf3 check with wrong password'):
            with patch.object(
                app_settings, 'IPERF3_CHECK_CONFIG', iperf3_conf_wrong_pass
            ), patch.object(Iperf3, 'schema', get_iperf3_schema()):
                self._set_auth_expected_calls(iperf3_conf_wrong_pass)
                mock_exec_command.side_effect = [
                    (RESULT_AUTH_FAIL, 1),
                    (RESULT_AUTH_FAIL, 1),
                ]

                result = self._perform_iperf3_check()
                self._assert_iperf3_fail_result(result)
                self.assertEqual(mock_warn.call_count, 2)
                self.assertEqual(mock_exec_command.call_count, 2)
                mock_warn.assert_has_calls(self._EXPECTED_WARN_CALLS)
                mock_exec_command.assert_has_calls(self._EXPECTED_COMMAND_CALLS)
            mock_warn.reset_mock()
            mock_exec_command.reset_mock()

        with self.subTest('Test iperf3 check with wrong username'):
            with patch.object(
                app_settings, 'IPERF3_CHECK_CONFIG', iperf3_conf_wrong_user
            ), patch.object(Iperf3, 'schema', get_iperf3_schema()):
                self._set_auth_expected_calls(iperf3_conf_wrong_user)
                mock_exec_command.side_effect = [
                    (RESULT_AUTH_FAIL, 1),
                    (RESULT_AUTH_FAIL, 1),
                ]

                result = self._perform_iperf3_check()
                self._assert_iperf3_fail_result(result)
                self.assertEqual(mock_warn.call_count, 2)
                self.assertEqual(mock_exec_command.call_count, 2)
                mock_warn.assert_has_calls(self._EXPECTED_WARN_CALLS)
                mock_exec_command.assert_has_calls(self._EXPECTED_COMMAND_CALLS)

    @patch.object(Ssh, 'exec_command')
    @patch.object(iperf3_logger, 'warning')
    @patch.object(iperf3_logger, 'info')
    @patch.object(cache, 'add')
    def test_iperf3_check_task_with_multiple_server_config(self, *args):
        mock_add = args[0]
        mock_info = args[1]
        mock_warn = args[2]
        mock_exec_command = args[3]
        org = self.device.organization
        iperf3_multiple_server_config = {
            self.org_id: {'host': self._IPERF3_TEST_MULTIPLE_SERVERS}
        }
        check = Check.objects.get(check_type=self._IPERF3)

        self._EXPECTED_COMMAND_CALLS_SERVER_1 = [
            call(
                (
                    f'iperf3 -c {self._IPERF3_TEST_MULTIPLE_SERVERS[0]} -p 5201 -t 10 --connect-timeout 1000 '
                    '-b 0 -l 128K -w 0 -P 1  -J'
                ),
                raise_unexpected_exit=False,
            ),
            call(
                (
                    f'iperf3 -c {self._IPERF3_TEST_MULTIPLE_SERVERS[0]} -p 5201 -t 10 --connect-timeout 1000 '
                    '-b 30M -l 0 -w 0 -P 1  -u -J'
                ),
                raise_unexpected_exit=False,
            ),
        ]
        self._EXPECTED_COMMAND_CALLS_SERVER_2 = [
            call(
                (
                    f'iperf3 -c {self._IPERF3_TEST_MULTIPLE_SERVERS[1]} -p 5201 -t 10 --connect-timeout 1000 '
                    '-b 0 -l 128K -w 0 -P 1  -J'
                ),
                raise_unexpected_exit=False,
            ),
            call(
                (
                    f'iperf3 -c {self._IPERF3_TEST_MULTIPLE_SERVERS[1]} -p 5201 -t 10 --connect-timeout 1000 '
                    '-b 30M -l 0 -w 0 -P 1  -u -J'
                ),
                raise_unexpected_exit=False,
            ),
        ]

        with patch.object(app_settings, 'IPERF3_CHECK_CONFIG', {}):
            with self.subTest('Test iperf3 check without config'):
                self._perform_iperf3_check()
                mock_warn.assert_called_with(
                    (
                        f'Iperf3 servers for organization "{org}" '
                        f'is not configured properly, iperf3 check skipped!'
                    )
                )
                self.assertEqual(mock_warn.call_count, 1)
            mock_warn.reset_mock()

        with patch.object(
            app_settings,
            'IPERF3_CHECK_CONFIG',
            {'invalid_org_uuid': {'host': self._IPERF3_TEST_SERVER, 'time': 10}},
        ):
            with self.subTest('Test iperf3 check with invalid config'):
                self._perform_iperf3_check()
                mock_warn.assert_called_with(
                    (
                        f'Iperf3 servers for organization "{org}" '
                        f'is not configured properly, iperf3 check skipped!'
                    )
                )
                self.assertEqual(mock_warn.call_count, 1)
            mock_warn.reset_mock()

        with patch.object(
            app_settings, 'IPERF3_CHECK_CONFIG', iperf3_multiple_server_config
        ):
            with self.subTest(
                'Test iperf3 check when all iperf3 servers are available'
            ):
                mock_add.return_value = True
                mock_exec_command.side_effect = [(RESULT_TCP, 0), (RESULT_UDP, 0)]
                self._perform_iperf3_check()
                self.assertEqual(mock_warn.call_count, 0)
                self.assertEqual(mock_add.call_count, 1)
                self.assertEqual(mock_exec_command.call_count, 2)
                mock_exec_command.assert_has_calls(
                    self._EXPECTED_COMMAND_CALLS_SERVER_1
                )
            mock_add.reset_mock()
            mock_warn.reset_mock()
            mock_exec_command.reset_mock()

            with self.subTest(
                'Test iperf3 check when single iperf3 server are available'
            ):
                mock_add.side_effect = [False, True]
                mock_exec_command.side_effect = [(RESULT_TCP, 0), (RESULT_UDP, 0)]
                self._perform_iperf3_check()
                self.assertEqual(mock_warn.call_count, 0)
                self.assertEqual(mock_add.call_count, 2)
                self.assertEqual(mock_exec_command.call_count, 2)
                mock_exec_command.assert_has_calls(
                    self._EXPECTED_COMMAND_CALLS_SERVER_2
                )
            mock_add.reset_mock()
            mock_warn.reset_mock()
            mock_exec_command.reset_mock()

            with self.subTest(
                'Test iperf3 check when all iperf3 servers are occupied initially'
            ):
                # If all available iperf3 servers are occupied initially,
                # then push the task back in the queue and acquire the iperf3
                # server only after completion of previous running checks
                mock_add.side_effect = [False, False, True]
                mock_exec_command.side_effect = [(RESULT_TCP, 0), (RESULT_UDP, 0)]
                self._perform_iperf3_check()
                mock_info.has_called_with(
                    (
                        f'At the moment, all available iperf3 servers of organization "{org}" '
                        f'are busy running checks, putting "{check}" back in the queue..'
                    )
                )
                self.assertEqual(mock_info.call_count, 4)
                self.assertEqual(mock_add.call_count, 3)
                self.assertEqual(mock_exec_command.call_count, 2)
                mock_exec_command.assert_has_calls(
                    self._EXPECTED_COMMAND_CALLS_SERVER_1
                )



from swapper import swappable_setting

from .base.models import AbstractCheck


class Check(AbstractCheck):
    class Meta(AbstractCheck.Meta):
        abstract = False
        swappable = swappable_setting('check', 'Check')



# Generated by Django 4.0.1 on 2022-01-10 11:27

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('check', '0005_create_config_applied'),
    ]

    operations = [
        migrations.RenameField(
            model_name='check',
            old_name='check',
            new_name='check_type',
        ),
    ]



from urllib.parse import urljoin

from django.apps import AppConfig
from django.conf import settings
from django.core.cache import cache
from django.db.models import Count
from django.db.models.signals import post_delete, post_save
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from swapper import get_model_name, load_model

from openwisp_controller.config.signals import (
    checksum_requested,
    config_status_changed,
    device_activated,
    device_deactivated,
)
from openwisp_controller.connection import settings as connection_settings
from openwisp_controller.connection.signals import is_working_changed
from openwisp_utils.admin_theme import (
    register_dashboard_chart,
    register_dashboard_template,
)
from openwisp_utils.admin_theme.menu import register_menu_subitem

from ..check import settings as check_settings
from ..monitoring.signals import threshold_crossed
from ..settings import MONITORING_API_BASEURL, MONITORING_API_URLCONF
from ..utils import transaction_on_commit
from . import settings as app_settings
from .signals import device_metrics_received, health_status_changed
from .utils import (
    get_device_cache_key,
    manage_default_retention_policy,
    manage_short_retention_policy,
)


class DeviceMonitoringConfig(AppConfig):
    name = 'openwisp_monitoring.device'
    label = 'device_monitoring'
    verbose_name = _('Device Monitoring')

    def ready(self):
        manage_default_retention_policy()
        manage_short_retention_policy()
        self.connect_is_working_changed()
        self.connect_device_signals()
        self.connect_config_status_changed()
        self.connect_wifi_client_signals()
        self.connect_offline_device_close_wifisession()
        self.connect_check_signals()
        self.device_recovery_detection()
        self.set_update_config_model()
        self.register_dashboard_items()
        self.register_menu_groups()
        self.add_connection_ignore_notification_reasons()

    def connect_check_signals(self):
        from django.db.models.signals import post_delete, post_save
        from swapper import load_model

        Check = load_model('check', 'Check')
        DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')

        post_save.connect(
            DeviceMonitoring.handle_critical_metric,
            sender=Check,
            dispatch_uid='check_post_save_receiver',
        )
        post_delete.connect(
            DeviceMonitoring.handle_critical_metric,
            sender=Check,
            dispatch_uid='check_post_delete_receiver',
        )

    def connect_device_signals(self):
        from .api.views import DeviceMetricView

        Device = load_model('config', 'Device')
        DeviceData = load_model('device_monitoring', 'DeviceData')
        DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
        DeviceLocation = load_model('geo', 'DeviceLocation')
        Metric = load_model('monitoring', 'Metric')
        Chart = load_model('monitoring', 'Chart')
        Organization = load_model('openwisp_users', 'Organization')

        post_save.connect(
            self.device_post_save_receiver,
            sender=Device,
            dispatch_uid='device_post_save_receiver',
        )
        post_save.connect(
            DeviceData.invalidate_cache,
            sender=Device,
            dispatch_uid='post_save_device_invalidate_devicedata_cache',
        )
        post_save.connect(
            DeviceData.invalidate_cache,
            sender=DeviceLocation,
            dispatch_uid='post_save_devicelocation_invalidate_devicedata_cache',
        )
        post_save.connect(
            DeviceMetricView.invalidate_get_charts_cache,
            sender=Metric,
            dispatch_uid=('metric_post_save_invalidate_view_charts_cache'),
        )
        post_save.connect(
            DeviceMetricView.invalidate_get_charts_cache,
            sender=Chart,
            dispatch_uid=('chart_post_save_invalidate_view_charts_cache'),
        )

        post_delete.connect(
            self.device_post_delete_receiver,
            sender=Device,
            dispatch_uid='device_post_delete_receiver',
        )
        post_delete.connect(
            DeviceMetricView.invalidate_get_device_cache,
            sender=Device,
            dispatch_uid=('device_post_delete_invalidate_view_device_cache'),
        )
        post_delete.connect(
            DeviceMetricView.invalidate_get_charts_cache,
            sender=Device,
            dispatch_uid=('device_post_delete_invalidate_view_charts_cache'),
        )
        post_delete.connect(
            DeviceMetricView.invalidate_get_charts_cache,
            sender=Metric,
            dispatch_uid=('metric_post_delete_invalidate_view_charts_cache'),
        )
        post_delete.connect(
            DeviceMetricView.invalidate_get_charts_cache,
            sender=Chart,
            dispatch_uid=('chart_post_delete_invalidate_view_charts_cache'),
        )
        post_delete.connect(
            DeviceData.invalidate_cache,
            sender=Device,
            dispatch_uid='post_delete_device_invalidate_devicedata_cache',
        )
        post_delete.connect(
            DeviceData.invalidate_cache,
            sender=DeviceLocation,
            dispatch_uid='post_delete_devicelocation_invalidate_devicedata_cache',
        )
        post_save.connect(
            self.organization_post_save_receiver,
            sender=Organization,
            dispatch_uid='post_save_organization_disabled_monitoring',
        )
        device_deactivated.connect(
            DeviceMonitoring.handle_deactivated_device,
            sender=Device,
            dispatch_uid='device_deactivated_update_devicemonitoring',
        )
        device_deactivated.connect(
            DeviceMetricView.invalidate_get_device_cache,
            sender=Device,
            dispatch_uid=('device_deactivated_invalidate_view_device_cache'),
        )
        device_activated.connect(
            DeviceMonitoring.handle_activated_device,
            sender=Device,
            dispatch_uid='device_activated_update_devicemonitoring',
        )
        device_activated.connect(
            DeviceMetricView.invalidate_get_device_cache,
            sender=Device,
            dispatch_uid=('device_activated_invalidate_view_device_cache'),
        )

    @classmethod
    def device_post_save_receiver(cls, instance, created, **kwargs):
        if created:
            DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
            DeviceMonitoring.objects.create(device=instance)

    @classmethod
    def device_post_delete_receiver(cls, instance, **kwargs):
        DeviceData = load_model('device_monitoring', 'DeviceData')
        instance.__class__ = DeviceData
        instance.checks.all().delete()
        instance.metrics.all().delete()

    @classmethod
    def organization_post_save_receiver(cls, instance, *args, **kwargs):
        if instance.is_active is False:
            from .tasks import handle_disabled_organization

            handle_disabled_organization.delay(str(instance.id))

    def device_recovery_detection(self):
        if not app_settings.DEVICE_RECOVERY_DETECTION:
            return

        Device = load_model('config', 'Device')
        DeviceData = load_model('device_monitoring', 'DeviceData')
        DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
        health_status_changed.connect(
            self.manage_device_recovery_cache_key,
            sender=DeviceMonitoring,
            dispatch_uid='recovery_health_status_changed',
        )
        checksum_requested.connect(
            self.trigger_device_recovery_checks,
            sender=Device,
            dispatch_uid='recovery_checksum_requested',
        )
        device_metrics_received.connect(
            self.trigger_device_recovery_checks,
            sender=DeviceData,
            dispatch_uid='recovery_device_metrics_received',
        )

    @classmethod
    def manage_device_recovery_cache_key(cls, instance, status, **kwargs):
        """Returns a cache key string.

        It sets the ``cache_key`` as 1 when device ``health_status`` goes
        to ``critical`` and deletes the ``cache_key`` when device recovers
        from ``critical`` state
        """
        cache_key = get_device_cache_key(device=instance.device)
        if status == 'critical':
            cache.set(cache_key, 1, timeout=None)
        else:
            cache.delete(cache_key)

    @classmethod
    def trigger_device_recovery_checks(cls, instance, **kwargs):
        from .tasks import trigger_device_checks

        # Cache is managed by "manage_device_recovery_cache_key".
        if cache.get(get_device_cache_key(device=instance), False):
            transaction_on_commit(lambda: trigger_device_checks.delay(pk=instance.pk))

    @classmethod
    def connect_is_working_changed(cls):
        is_working_changed.connect(
            cls.is_working_changed_receiver,
            sender=load_model('connection', 'DeviceConnection'),
            dispatch_uid='is_working_changed_monitoring',
        )

    @classmethod
    def is_working_changed_receiver(
        cls, instance, is_working, old_is_working, failure_reason, **kwargs
    ):
        from .tasks import trigger_device_checks

        Check = load_model('check', 'Check')
        device = instance.device
        device_monitoring = device.monitoring
        # if old_is_working is None, it's a new device connection which wasn't
        # ever used yet, so nothing is really changing and we don't need to do anything
        if old_is_working is None and is_working:
            return
        # if device is down because of connectivity issues, it's probably due
        # to reboot caused by firmware upgrade, avoid notifications
        ignored_failures = ['Unable to connect', 'timed out']
        for message in ignored_failures:
            if message in failure_reason:
                device_monitoring.save()
                return
        initial_status = device_monitoring.status
        status = 'ok' if is_working else 'problem'
        # do not send notifications if recovery made after firmware upgrade
        if status == initial_status == 'ok':
            device_monitoring.save()
            return
        if not is_working:
            if initial_status == 'ok':
                transaction_on_commit(lambda: trigger_device_checks.delay(pk=device.pk))
        else:
            # if checks exist trigger them else, set status as 'ok'
            if Check.objects.filter(object_id=instance.device.pk).exists():
                transaction_on_commit(lambda: trigger_device_checks.delay(pk=device.pk))
            else:
                device_monitoring.update_status(status)

    @classmethod
    def connect_config_status_changed(cls):
        Config = load_model('config', 'Config')

        if check_settings.AUTO_CONFIG_CHECK:
            config_status_changed.connect(
                cls.config_status_changed_receiver,
                sender=Config,
                dispatch_uid='monitoring.config_status_changed_receiver',
            )

    @classmethod
    def connect_wifi_client_signals(cls):
        if not app_settings.WIFI_SESSIONS_ENABLED:
            return
        WifiClient = load_model('device_monitoring', 'WifiClient')
        post_save.connect(
            WifiClient.invalidate_cache,
            sender=WifiClient,
            dispatch_uid='post_save_invalidate_wificlient_cache',
        )
        post_delete.connect(
            WifiClient.invalidate_cache,
            sender=WifiClient,
            dispatch_uid='post_delete_invalidate_wificlient_cache',
        )

    @classmethod
    def connect_offline_device_close_wifisession(cls):
        if not app_settings.WIFI_SESSIONS_ENABLED:
            return

        Metric = load_model('monitoring', 'Metric')
        WifiSession = load_model('device_monitoring', 'WifiSession')
        threshold_crossed.connect(
            WifiSession.offline_device_close_session,
            sender=Metric,
            dispatch_uid='offline_device_close_session',
        )

    @classmethod
    def config_status_changed_receiver(cls, sender, instance, **kwargs):
        from ..check.tasks import perform_check

        DeviceData = load_model('device_monitoring', 'DeviceData')
        device = DeviceData.objects.get(config=instance)
        check = device.checks.filter(check_type__contains='ConfigApplied').first()
        if check:
            transaction_on_commit(lambda: perform_check.delay(check.pk))

    def set_update_config_model(self):
        if not getattr(settings, 'OPENWISP_UPDATE_CONFIG_MODEL', None):
            setattr(
                connection_settings,
                'UPDATE_CONFIG_MODEL',
                'device_monitoring.DeviceData',
            )

    def register_dashboard_items(self):
        WifiSession = load_model('device_monitoring', 'WifiSession')
        Chart = load_model('monitoring', 'Chart')
        register_dashboard_chart(
            position=0,
            config={
                'name': _('Monitoring Status'),
                'query_params': {
                    'app_label': 'config',
                    'model': 'device',
                    'group_by': 'monitoring__status',
                },
                'colors': {
                    'ok': '#267126',
                    'problem': '#ffb442',
                    'critical': '#a72d1d',
                    'unknown': '#353c44',
                    'deactivated': '#000',
                },
                'labels': {
                    'ok': app_settings.HEALTH_STATUS_LABELS['ok'],
                    'problem': app_settings.HEALTH_STATUS_LABELS['problem'],
                    'critical': app_settings.HEALTH_STATUS_LABELS['critical'],
                    'unknown': app_settings.HEALTH_STATUS_LABELS['unknown'],
                    'deactivated': app_settings.HEALTH_STATUS_LABELS['deactivated'],
                },
            },
        )

        if app_settings.DASHBOARD_MAP:
            loc_geojson_url = reverse_lazy(
                'monitoring:api_location_geojson', urlconf=MONITORING_API_URLCONF
            )
            device_list_url = reverse_lazy(
                'monitoring:api_location_device_list',
                urlconf=MONITORING_API_URLCONF,
                args=['000'],
            )
            if MONITORING_API_BASEURL:
                device_list_url = urljoin(MONITORING_API_BASEURL, str(device_list_url))
                loc_geojson_url = urljoin(MONITORING_API_BASEURL, str(loc_geojson_url))

            register_dashboard_template(
                position=0,
                config={
                    'template': 'admin/dashboard/device_map.html',
                    'css': (
                        'monitoring/css/device-map.css',
                        'leaflet/leaflet.css',
                        'monitoring/css/leaflet.fullscreen.css',
                        'monitoring/css/netjsongraph.css',
                    ),
                    'js': (
                        'monitoring/js/lib/netjsongraph.min.js',
                        'monitoring/js/lib/leaflet.fullscreen.min.js',
                        'monitoring/js/device-map.js',
                    ),
                },
                extra_config={
                    'monitoring_device_list_url': device_list_url,
                    'monitoring_location_geojson_url': loc_geojson_url,
                },
            )

        general_wifi_client_quick_links = {}
        if app_settings.WIFI_SESSIONS_ENABLED:
            register_dashboard_chart(
                position=13,
                config={
                    'name': _('Currently Active WiFi Sessions'),
                    'query_params': {
                        'app_label': WifiSession._meta.app_label,
                        'model': WifiSession._meta.model_name,
                        'filter': {'stop_time__isnull': True},
                        'aggregate': {
                            'active__count': Count('id'),
                        },
                        'organization_field': 'device__organization_id',
                    },
                    'filters': {
                        'key': 'stop_time__isnull',
                        'active__count': 'true',
                    },
                    'colors': {
                        'active__count': '#267126',
                    },
                    'labels': {
                        'active__count': _('Currently Active WiFi Sessions'),
                    },
                    'quick_link': {
                        'url': reverse_lazy(
                            'admin:{app_label}_{model_name}_changelist'.format(
                                app_label=WifiSession._meta.app_label,
                                model_name=WifiSession._meta.model_name,
                            )
                        ),
                        'label': _('Open WiFi session list'),
                        'title': _('View full history of WiFi Sessions'),
                        'custom_css_classes': ['negative-top-20'],
                    },
                },
            )

            general_wifi_client_quick_links = {
                'General WiFi Clients': {
                    'url': reverse_lazy(
                        'admin:{app_label}_{model_name}_changelist'.format(
                            app_label=WifiSession._meta.app_label,
                            model_name=WifiSession._meta.model_name,
                        )
                    ),
                    'label': _('Open WiFi session list'),
                    'title': _('View full history of WiFi Sessions'),
                }
            }

        dashboard_chart_url = reverse_lazy(
            'monitoring_general:api_dashboard_timeseries',
            urlconf=MONITORING_API_URLCONF,
        )
        if MONITORING_API_BASEURL:
            dashboard_chart_url = urljoin(
                MONITORING_API_BASEURL, str(dashboard_chart_url)
            )
        register_dashboard_template(
            position=55,
            config={
                'template': 'monitoring/paritals/chart.html',
                'css': (
                    'monitoring/css/daterangepicker.css',
                    'monitoring/css/percircle.min.css',
                    'monitoring/css/chart.css',
                    'monitoring/css/dashboard-chart.css',
                    'admin/css/vendor/select2/select2.min.css',
                    'admin/css/autocomplete.css',
                ),
                'js': (
                    'admin/js/vendor/select2/select2.full.min.js',
                    'monitoring/js/lib/moment.min.js',
                    'monitoring/js/lib/daterangepicker.min.js',
                    'monitoring/js/lib/percircle.min.js',
                    'monitoring/js/chart.js',
                    'monitoring/js/chart-utils.js',
                    'monitoring/js/dashboard-chart.js',
                ),
            },
            extra_config={
                'api_url': dashboard_chart_url,
                'default_time': Chart.DEFAULT_TIME,
                'chart_quick_links': general_wifi_client_quick_links,
            },
            after_charts=True,
        )

    def register_menu_groups(self):
        if app_settings.WIFI_SESSIONS_ENABLED:
            register_menu_subitem(
                group_position=80,
                item_position=0,
                config={
                    'label': _('WiFi Sessions'),
                    'model': get_model_name('device_monitoring', 'WifiSession'),
                    'name': 'changelist',
                    'icon': 'ow-monitoring-wifi',
                },
            )

    def add_connection_ignore_notification_reasons(self):
        from openwisp_controller.connection.apps import ConnectionConfig

        ConnectionConfig._ignore_connection_notification_reasons.extend(
            ['timed out', 'Unable to connect']
        )



from django.apps import AppConfig
from django.db.models.signals import post_save
from django.utils.translation import gettext_lazy as _
from swapper import load_model

from openwisp_monitoring.check import settings as app_settings


class CheckConfig(AppConfig):
    name = 'openwisp_monitoring.check'
    label = 'check'
    verbose_name = _('Network Monitoring Checks')

    def ready(self):
        self._connect_signals()

    def _connect_signals(self):
        if app_settings.AUTO_PING:
            from .base.models import auto_ping_receiver

            post_save.connect(
                auto_ping_receiver,
                sender=load_model('config', 'Device'),
                dispatch_uid='auto_ping',
            )

        if app_settings.AUTO_CONFIG_CHECK:
            from .base.models import auto_config_check_receiver

            post_save.connect(
                auto_config_check_receiver,
                sender=load_model('config', 'Device'),
                dispatch_uid='auto_config_check',
            )
        if app_settings.AUTO_IPERF3:
            from .base.models import auto_iperf3_check_receiver

            post_save.connect(
                auto_iperf3_check_receiver,
                sender=load_model('config', 'Device'),
                dispatch_uid='auto_iperf3_check',
            )



import logging

from celery import shared_task
from django.core.exceptions import ObjectDoesNotExist
from django.utils.timezone import now, timedelta
from swapper import load_model

from openwisp_utils.tasks import OpenwispCeleryTask

from ..check.tasks import perform_check

logger = logging.getLogger(__name__)


@shared_task(base=OpenwispCeleryTask)
def trigger_device_checks(pk, recovery=True):
    """Triggers the monitoring checks for the specified device pk.

    Retrieves all related checks to the passed ``device`` and calls the
    ``perform_check`` task from each of them.

    If no check exists changes the status according to the ``recovery``
    argument.
    """
    DeviceData = load_model('device_monitoring', 'DeviceData')
    try:
        device = DeviceData.objects.get(pk=pk)
    except ObjectDoesNotExist:
        logger.warning(f'The device with uuid {pk} has been deleted')
        return
    checks = device.checks.filter(is_active=True).only('id').values('id')
    has_checks = False
    for check in checks:
        perform_check.delay(check['id'])
        has_checks = True
    if not has_checks:
        status = 'ok' if recovery else 'critical'
        device.monitoring.update_status(status)


@shared_task(base=OpenwispCeleryTask)
def delete_wifi_clients_and_sessions(days=6 * 30):
    WifiClient = load_model('device_monitoring', 'WifiClient')
    WifiSession = load_model('device_monitoring', 'WifiSession')

    WifiSession.objects.filter(start_time__lte=(now() - timedelta(days=days))).delete()
    WifiClient.objects.exclude(
        mac_address__in=WifiSession.objects.values_list('wifi_client')
    ).delete()


@shared_task(base=OpenwispCeleryTask)
def offline_device_close_session(device_id):
    WifiSession = load_model('device_monitoring', 'WifiSession')
    WifiSession.objects.filter(device_id=device_id, stop_time__isnull=True).update(
        stop_time=now()
    )


@shared_task(base=OpenwispCeleryTask)
def write_device_metrics(pk, data, time=None, current=False):
    DeviceData = load_model('device_monitoring', 'DeviceData')
    try:
        device_data = DeviceData.get_devicedata(str(pk))
    except DeviceData.DoesNotExist:
        return
    device_data.writer.write(data, time, current)


@shared_task(base=OpenwispCeleryTask)
def handle_disabled_organization(organization_id):
    DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
    DeviceMonitoring.handle_disabled_organization(organization_id)



from datetime import datetime, timedelta
from unittest.mock import patch

from celery.exceptions import Retry
from django.core.exceptions import ValidationError
from django.test import TestCase, tag
from django.utils.timezone import make_aware, now
from freezegun import freeze_time
from influxdb import InfluxDBClient
from influxdb.exceptions import InfluxDBClientError, InfluxDBServerError
from pytz import timezone as tz
from swapper import load_model

from openwisp_monitoring.device.settings import (
    DEFAULT_RETENTION_POLICY,
    SHORT_RETENTION_POLICY,
)
from openwisp_monitoring.device.utils import (
    DEFAULT_RP,
    SHORT_RP,
    manage_default_retention_policy,
    manage_short_retention_policy,
)
from openwisp_monitoring.monitoring.tests import TestMonitoringMixin
from openwisp_monitoring.settings import MONITORING_TIMESERIES_RETRY_OPTIONS
from openwisp_utils.tests import capture_stderr

from ...exceptions import TimeseriesWriteException
from .. import timeseries_db

Chart = load_model('monitoring', 'Chart')
Notification = load_model('openwisp_notifications', 'Notification')


@tag('timeseries_client')
class TestDatabaseClient(TestMonitoringMixin, TestCase):
    def test_forbidden_queries(self):
        queries = [
            'DROP DATABASE openwisp2',
            'DROP MEASUREMENT test_metric',
            'CREATE DATABASE test',
            'DELETE MEASUREMENT test_metric',
            'ALTER RETENTION POLICY policy',
            'SELECT * INTO metric2 FROM test_metric',
        ]
        for q in queries:
            try:
                timeseries_db.validate_query(q)
            except ValidationError as e:
                self.assertIn('configuration', e.message_dict)
            else:
                self.fail('ValidationError not raised')

    def test_get_custom_query(self):
        c = self._create_chart(test_data=None)
        custom_q = c._default_query.replace('{field_name}', '{fields}')
        q = c.get_query(query=custom_q, fields=['SUM(*)'])
        self.assertIn('SELECT SUM(*) FROM', q)

    def test_is_aggregate_bug(self):
        m = self._create_object_metric(name='summary_avg')
        c = Chart(metric=m, configuration='dummy')
        self.assertFalse(timeseries_db._is_aggregate(c.query))

    def test_is_aggregate_fields_function(self):
        m = self._create_object_metric(name='is_aggregate_func')
        c = Chart(metric=m, configuration='uptime')
        self.assertTrue(timeseries_db._is_aggregate(c.query))

    def test_get_query_fields_function(self):
        c = self._create_chart(test_data=None, configuration='histogram')
        q = c.get_query(fields=['ssh', 'http2', 'apple-music'])
        expected = (
            'SELECT SUM("ssh") / 1 AS ssh, '
            'SUM("http2") / 1 AS http2, '
            'SUM("apple-music") / 1 AS apple_music FROM'
        )
        self.assertIn(expected, q)

    def test_default_query(self):
        c = self._create_chart(test_data=False)
        q = (
            "SELECT {field_name} FROM {key} WHERE time >= '{time}' {end_date} AND "
            "content_type = '{content_type}' AND object_id = '{object_id}'"
        )
        self.assertEqual(c.query, q)

    def test_write(self):
        timeseries_db.write('test_write', dict(value=2), database=self.TEST_DB)
        measurement = list(
            timeseries_db.query(
                'select * from test_write', database=self.TEST_DB
            ).get_points()
        )[0]
        self.assertEqual(measurement['value'], 2)

    def test_general_write(self):
        m = self._create_general_metric(name='Sync test')
        m.write(1)
        measurement = list(timeseries_db.query('select * from sync_test').get_points())[
            0
        ]
        self.assertEqual(measurement['value'], 1)

    def test_object_write(self):
        om = self._create_object_metric()
        om.write(3)
        content_type = '.'.join(om.content_type.natural_key())
        q = (
            f"select * from test_metric WHERE object_id = '{om.object_id}'"
            f" AND content_type = '{content_type}'"
        )
        measurement = timeseries_db.get_list_query(q)[0]
        self.assertEqual(measurement['value'], 3)

    def test_general_same_key_different_fields(self):
        down = self._create_general_metric(
            name='traffic (download)', key='traffic', field_name='download'
        )
        down.write(200)
        up = self._create_general_metric(
            name='traffic (upload)', key='traffic', field_name='upload'
        )
        up.write(100)
        measurement = list(
            timeseries_db.query('select download from traffic').get_points()
        )[0]
        self.assertEqual(measurement['download'], 200)
        measurement = list(
            timeseries_db.query('select upload from traffic').get_points()
        )[0]
        self.assertEqual(measurement['upload'], 100)

    def test_object_same_key_different_fields(self):
        user = self._create_user()
        user_down = self._create_object_metric(
            name='traffic (download)',
            key='traffic',
            field_name='download',
            content_object=user,
        )
        user_down.write(200)
        user_up = self._create_object_metric(
            name='traffic (upload)',
            key='traffic',
            field_name='upload',
            content_object=user,
        )
        user_up.write(100)
        content_type = '.'.join(user_down.content_type.natural_key())
        q = (
            f"select download from traffic WHERE object_id = '{user_down.object_id}'"
            f" AND content_type = '{content_type}'"
        )
        measurement = timeseries_db.get_list_query(q)[0]
        self.assertEqual(measurement['download'], 200)
        q = (
            f"select upload from traffic WHERE object_id = '{user_up.object_id}'"
            f" AND content_type = '{content_type}'"
        )
        measurement = timeseries_db.get_list_query(q)[0]
        self.assertEqual(measurement['upload'], 100)

    def test_delete_metric_data(self):
        m = self._create_general_metric(name='test_metric')
        m.write(100)
        self.assertEqual(m.read()[0]['value'], 100)
        timeseries_db.delete_metric_data(key=m.key)
        self.assertEqual(m.read(), [])
        om = self._create_object_metric(name='dummy')
        om.write(50)
        m.write(100)
        self.assertEqual(m.read()[0]['value'], 100)
        self.assertEqual(om.read()[0]['value'], 50)
        timeseries_db.delete_metric_data()
        self.assertEqual(m.read(), [])
        self.assertEqual(om.read(), [])

    def test_get_query_1d(self):
        c = self._create_chart(test_data=None, configuration='uptime')
        q = c.get_query(time='1d')
        last24 = now() - timedelta(days=1)
        self.assertIn(str(last24)[0:14], q)
        self.assertIn('group by time(10m)', q.lower())

    def test_get_query_30d(self):
        c = self._create_chart(test_data=None, configuration='uptime')
        q = c.get_query(time='30d')
        last30d = now() - timedelta(days=30)
        self.assertIn(str(last30d)[0:10], q)
        self.assertIn('group by time(24h)', q.lower())

    def test_group_by_tags(self):
        self.assertEqual(
            timeseries_db._group_by(
                'SELECT COUNT(item) FROM measurement GROUP BY time(1d)',
                time='30d',
                chart_type='stackedbar+lines',
                group_map={'30d': '30d'},
                strip=False,
            ),
            'SELECT COUNT(item) FROM measurement GROUP BY time(30d)',
        )
        self.assertEqual(
            timeseries_db._group_by(
                'SELECT COUNT(item) FROM measurement GROUP BY time(1d)',
                time='30d',
                chart_type='stackedbar+lines',
                group_map={'30d': '30d'},
                strip=True,
            ),
            'SELECT COUNT(item) FROM measurement ',
        )
        self.assertEqual(
            timeseries_db._group_by(
                'SELECT COUNT(item) FROM measurement GROUP BY time(1d), tag',
                time='30d',
                chart_type='stackedbar+lines',
                group_map={'30d': '30d'},
                strip=False,
            ),
            'SELECT COUNT(item) FROM measurement GROUP BY time(30d), tag',
        )
        self.assertEqual(
            timeseries_db._group_by(
                'SELECT COUNT(item) FROM measurement GROUP BY time(1d), tag',
                time='30d',
                chart_type='stackedbar+lines',
                group_map={'30d': '30d'},
                strip=True,
            ),
            'SELECT COUNT(item) FROM measurement GROUP BY tag',
        )

    def test_retention_policy(self):
        manage_short_retention_policy()
        manage_default_retention_policy()
        rp = timeseries_db.get_list_retention_policies()
        self.assertEqual(len(rp), 2)
        self.assertEqual(rp[0]['name'], DEFAULT_RP)
        self.assertEqual(rp[0]['default'], True)
        self.assertEqual(rp[0]['duration'], DEFAULT_RETENTION_POLICY)
        self.assertEqual(rp[1]['name'], SHORT_RP)
        self.assertEqual(rp[1]['default'], False)
        self.assertEqual(rp[1]['duration'], SHORT_RETENTION_POLICY)

    def test_query_set(self):
        c = self._create_chart(configuration='histogram')
        expected = (
            "SELECT {fields|SUM|/ 1} FROM {key} "
            "WHERE time >= '{time}' AND content_type = "
            "'{content_type}' AND object_id = '{object_id}'"
        )
        self.assertEqual(c.query, expected)
        self.assertEqual(
            ''.join(timeseries_db.queries.default_chart_query[0:2]), c._default_query
        )
        c.metric.object_id = None
        self.assertEqual(timeseries_db.queries.default_chart_query[0], c._default_query)

    def test_read_order(self):
        m = self._create_general_metric(name='dummy')
        m.write(30)
        m.write(40, time=now() - timedelta(days=2))
        with self.subTest('Test ascending read order'):
            metric_data = m.read(limit=2, order='time')
            self.assertEqual(metric_data[0]['value'], 40)
            self.assertEqual(metric_data[1]['value'], 30)
        with self.subTest('Test descending read order'):
            metric_data = m.read(limit=2, order='-time')
            self.assertEqual(metric_data[0]['value'], 30)
            self.assertEqual(metric_data[1]['value'], 40)
        with self.subTest('Test invalid read order'):
            with self.assertRaises(timeseries_db.client_error) as e:
                metric_data = m.read(limit=2, order='invalid')
                self.assertIn('Invalid order "invalid" passed.', str(e))

    def test_read_with_rp(self):
        self._create_admin()
        manage_short_retention_policy()
        with self.subTest(
            'Test metric write on short retention_policy immediate alert'
        ):
            m = self._create_general_metric(name='dummy')
            self._create_alert_settings(
                metric=m, custom_operator='<', custom_threshold=1, custom_tolerance=0
            )
            m.write(0, retention_policy=SHORT_RP)
            self.assertEqual(m.read(retention_policy=SHORT_RP)[0][m.field_name], 0)
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)
        with self.subTest(
            'Test metric write on short retention_policy with deferred alert'
        ):
            m2 = self._create_general_metric(name='dummy2')
            self._create_alert_settings(
                metric=m2, custom_operator='<', custom_threshold=1, custom_tolerance=1
            )
            m.write(0, retention_policy=SHORT_RP, time=now() - timedelta(minutes=2))
            self.assertEqual(m.read(retention_policy=SHORT_RP)[0][m.field_name], 0)
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)

    def test_metric_write_microseconds_precision(self):
        m = self._create_object_metric(
            name='wlan0', key='wlan0', configuration='clients'
        )
        m.write(
            "00:14:5c:00:00:00",
            time=make_aware(datetime(2020, 7, 31, 22, 5, 47, 235142)),
        )
        m.write(
            "00:23:4a:00:00:00",
            time=make_aware(datetime(2020, 7, 31, 22, 5, 47, 235152)),
        )
        self.assertEqual(len(m.read()), 2)

    @patch.object(
        InfluxDBClient, 'write', side_effect=InfluxDBServerError('Server error')
    )
    @capture_stderr()
    def test_write_retry(self, mock_write):
        with self.assertRaises(TimeseriesWriteException):
            timeseries_db.write('test_write', {'value': 1})
        m = self._create_general_metric(name='Test metric')
        with self.assertRaises(Retry):
            m.write(1)

    @patch.object(
        InfluxDBClient,
        'write',
        side_effect=InfluxDBClientError(
            content='{"error":"partial write: points beyond retention policy dropped=1"}',
            code=400,
        ),
    )
    @capture_stderr()
    def test_write_skip_retry_for_retention_policy(self, mock_write):
        try:
            timeseries_db.write('test_write', {'value': 1})
        except TimeseriesWriteException:
            self.fail(
                'TimeseriesWriteException should not be raised when data '
                'points crosses retention policy'
            )
        m = self._create_general_metric(name='Test metric')
        try:
            m.write(1)
        except Retry:
            self.fail(
                'Writing metric should not be retried when data '
                'points crosses retention policy'
            )

    @patch.object(
        InfluxDBClient, 'write', side_effect=InfluxDBServerError('Server error')
    )
    @capture_stderr()
    def test_timeseries_write_params(self, mock_write):
        with freeze_time('Jan 14th, 2020') as frozen_datetime:
            m = self._create_general_metric(name='Test metric')
            with self.assertRaises(Retry) as e:
                m.write(1)
            frozen_datetime.tick(delta=timedelta(minutes=10))
            self.assertEqual(
                now(), datetime(2020, 1, 14, tzinfo=tz('UTC')) + timedelta(minutes=10)
            )
            task_signature = e.exception.sig
            with patch.object(timeseries_db, 'write') as mock_write:
                self._retry_task(task_signature)
            mock_write.assert_called_with(
                'test_metric',
                {'value': 1},
                database=None,
                retention_policy=None,
                tags={},
                # this should be the original time at the moment of first failure
                timestamp=datetime(2020, 1, 14, tzinfo=tz('UTC')).isoformat(),
                current=False,
            )

    def _retry_task(self, task_signature):
        task_kwargs = task_signature.kwargs
        task_signature.type.run(**task_kwargs)

    @patch.object(
        InfluxDBClient, 'query', side_effect=InfluxDBServerError('Server error')
    )
    def test_retry_mechanism(self, mock_query):
        max_retries = MONITORING_TIMESERIES_RETRY_OPTIONS.get('max_retries')
        with patch('logging.Logger.info') as mocked_logger:
            try:
                self.test_get_query_fields_function()
            except Exception:
                pass
            self.assertEqual(mocked_logger.call_count, max_retries)
            mocked_logger.assert_called_with(
                'Error while executing method "query":\nServer error\nAttempt '
                f'{max_retries} out of {max_retries}.\n'
            )


class TestDatabaseClientUdp(TestMonitoringMixin, TestCase):
    def test_exceed_udp_packet_limit(self):
        # When using UDP to write data to InfluxDB, writing
        # huge data that exceeds UDP packet limit should not raise
        # an error. Instead, the client should fallback to the
        # TCP connection.
        timeseries_db.write(
            'test_udp_write', dict(value='O' * 66000), database=self.TEST_DB
        )
        measurement = list(
            timeseries_db.query(
                'select * from test_udp_write', database=self.TEST_DB
            ).get_points()
        )
        self.assertEqual(len(measurement), 1)
